"""
Misc tool collection
"""
import importlib
import inspect
import threading
from typing import List, Callable, Tuple, TypeVar, Any, Type, Optional, Union, Iterable

from mcdreforged.minecraft.rtext import RTextBase


def start_thread(func: Callable, args: Tuple, name: Optional[str] = None):
	thread = threading.Thread(target=func, args=args, name=name, daemon=True)
	thread.start()
	return thread


def load_class(path: str):
	"""
	:param path: the path to the class, e.g. mcdreforged.info_reactor.info.Info
	:return: The class
	"""
	try:
		module_path, class_name = path.rsplit('.', 1)
	except ValueError:
		raise ImportError('Wrong path to a class: {}'.format(path)) from None
	module = importlib.import_module(module_path)
	try:
		return getattr(module, class_name)
	except AttributeError:
		raise ImportError('Class "{}" not found in package "{}"'.format(class_name, module_path)) from None


T = TypeVar('T')


def unique_list(lst: List[T]) -> List[T]:
	ret = list(set(lst))
	ret.sort(key=lst.index)
	return ret


def get_all_base_class(cls):
	if cls is object:
		return []
	ret = [cls]
	for base in cls.__bases__:
		ret.extend(get_all_base_class(base))
	return unique_list(ret)


def print_text_to_console(logger, text):
	for line in RTextBase.from_any(text).to_colored_text().splitlines():
		logger.info(line)


def check_type(value: Any, types: Union[Type, Iterable[Type]], error_message: str = None):
	if not isinstance(types, Iterable):
		types = [types]
	if not any(map(lambda t: isinstance(value, t), types)):
		if error_message is None:
			error_message = 'Except type {} but found type {}'.format(types, type(value))
		raise TypeError(error_message)


def check_class(class_: Type, base_class: Type, error_message: str = None):
	if not issubclass(class_, base_class):
		if error_message is None:
			error_message = 'Except class derived from {} but found class {}'.format(base_class, class_)
		raise TypeError(error_message)


def copy_signature(target: Callable, origin: Callable) -> Callable:
	"""
	Copy the function signature of origin into target
	"""
	# https://stackoverflow.com/questions/39926567/python-create-decorator-preserving-function-arguments
	target.__signature__ = inspect.signature(origin)
	return target


def represent(obj: Any) -> str:
	"""
	aka repr
	"""
	return '{}[{}]'.format(type(obj).__name__, ','.join([
		'{}={}'.format(k, repr(v)) for k, v in vars(obj).items() if not k.startswith('_')
	]))


class WaitableCallable:
	def __init__(self, func: Callable):
		self.__func = func
		self.__event = threading.Event()

	def __call__(self, *args, **kwargs):
		rv = self.__func(*args, **kwargs)
		self.__event.set()
		return rv

	def wait(self, timeout: Optional[float] = None) -> bool:
		"""
		Return if the event has been set
		"""
		return self.__event.wait(timeout=timeout)


import locale
import os
import sys
import time
import traceback
from subprocess import Popen, PIPE, STDOUT
from threading import Lock
from typing import Optional, Callable, Any

import pkg_resources
import psutil
from ruamel.yaml import YAMLError

from mcdreforged.command.command_manager import CommandManager
from mcdreforged.config import Config
from mcdreforged.constants import core_constant
from mcdreforged.executor.console_handler import ConsoleHandler
from mcdreforged.executor.task_executor import TaskExecutor
from mcdreforged.executor.update_helper import UpdateHelper
from mcdreforged.executor.watchdog import WatchDog
from mcdreforged.handler.server_handler_manager import ServerHandlerManager
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.info_reactor_manager import InfoReactorManager
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.mcdr_state import ServerState, MCDReforgedState, MCDReforgedFlag
from mcdreforged.minecraft.rcon.rcon_manager import RconManager
from mcdreforged.permission.permission_manager import PermissionManager
from mcdreforged.plugin.plugin_event import MCDRPluginEvents
from mcdreforged.plugin.plugin_manager import PluginManager
from mcdreforged.plugin.server_interface import ServerInterface
from mcdreforged.preference.preference_manager import PreferenceManager
from mcdreforged.translation.translation_manager import TranslationManager
from mcdreforged.utils import file_util
from mcdreforged.utils.exception import IllegalCallError, ServerStopped, ServerStartError, IllegalStateError
from mcdreforged.utils.logger import DebugOption, MCDReforgedLogger, MCColoredFormatter
from mcdreforged.utils.types import MessageText


class MCDReforgedServer:
	process: Optional[Popen]

	def __init__(self, *, generate_default_only: bool = False, initialize_environment: bool = False):
		"""
		:param generate_default_only: If set to true, MCDR will only generate the default configure and permission files
		"""
		self.mcdr_state = MCDReforgedState.INITIALIZING
		self.server_state = ServerState.STOPPED
		self.server_information = ServerInformation()
		self.process = None  # type: Optional[PIPE]
		self.flags = MCDReforgedFlag.NONE
		self.starting_server_lock = Lock()  # to prevent multiple start_server() call
		self.stop_lock = Lock()  # to prevent multiple stop() call

		# will be assigned in on_config_changed()
		self.encoding_method = None  # type: Optional[str]
		self.decoding_method = None  # type: Optional[str]

		# --- Constructing fields --- #
		self.logger = MCDReforgedLogger()
		self.config = Config(self.logger)
		self.permission_manager = PermissionManager(self)
		self.basic_server_interface = ServerInterface(self)
		self.task_executor = TaskExecutor(self)
		self.console_handler = ConsoleHandler(self)
		self.watch_dog = WatchDog(self)
		self.update_helper = UpdateHelper(self)
		self.translation_manager = TranslationManager(self.logger)
		self.rcon_manager = RconManager(self)
		self.server_handler_manager = ServerHandlerManager(self)
		self.reactor_manager = InfoReactorManager(self)
		self.command_manager = CommandManager(self)
		self.plugin_manager = PluginManager(self)
		self.preference_manager = PreferenceManager(self)

		self.__check_environment()

		# --- Input arguments "generate_default_only" processing --- #
		if generate_default_only:
			self.config.save_default()
			self.permission_manager.save_default()
			return

		# --- Initialize fields instance --- #
		self.translation_manager.load_translations()  # translations are used for logging, so load them first
		if initialize_environment:
			# Prepare config / permission files if they're missing
			if not self.config.file_presents():
				self.config.save_default()
				default_config = self.config.get_default_yaml()
				file_util.touch_directory(default_config['working_directory'])  # create server/ folder
			if not self.permission_manager.file_presents():
				self.permission_manager.save_default()

		# Check if there's any file missing
		# If there's any, MCDR environment might not be probably setup
		file_missing = False

		def load(kind: str, func: Callable[[], Any]) -> bool:
			nonlocal file_missing
			try:
				func()
			except FileNotFoundError:
				self.logger.error('{} is missing'.format(kind.title()))
				file_missing = True
			except YAMLError as e:
				self.logger.error('Failed to load {}: {}'.format(kind, type(e).__name__))
				for line in str(e).splitlines():
					self.logger.error(line)
				return False
			else:
				return True

		# load_config: config, language, handlers, plugin directories, reactors, handlers
		# load_permission_file: permission
		# config change will lead to creating plugin folders
		loading_success = \
			load('configure', lambda: self.load_config(allowed_missing_file=False, echo=not initialize_environment)) and \
			load('permission', lambda: self.permission_manager.load_permission_file(allowed_missing_file=False))
		if file_missing:
			self.on_file_missing()
			return
		if not loading_success:
			return

		# MCDR environment has been setup, so continue creating default folders and loading stuffs
		self.logger.set_file(core_constant.LOGGING_FILE)  # will create logs/ folder
		self.plugin_manager.touch_directory()  # will create config/ folder

		# --- Done --- #
		self.set_mcdr_state(MCDReforgedState.INITIALIZED)

	def __del__(self):
		try:
			if self.process and self.process.poll() is None:
				self.__kill_server()
		except:
			pass

	def __check_environment(self):
		"""
		Some checks at initialization
		In dev environment, you can use setup.py to create `mcdreforged.egg-info/` so the package check will pass
		"""
		mcdr_pkg = core_constant.PACKAGE_NAME  # should be "mcdreforged"
		try:
			pkg_resources.require(mcdr_pkg)
		except pkg_resources.ResolutionError:
			self.logger.warning('It looks like you\'re launching from source as {} is not found in python packages'.format(mcdr_pkg))
			self.logger.warning('In this way, the plugin system might not work correctly')
			self.logger.warning('In a production environment, you should install {} from PyPI, see document ({}) for more information'.format(mcdr_pkg, core_constant.DOCUMENTION_URL))
			self.logger.warning('MCDR will launch after 10 seconds...')
			time.sleep(10)

	def on_file_missing(self):
		self.logger.info('Looks like MCDR is not initialized at current directory {}'.format(os.getcwd()))
		self.logger.info('Use "python -m {} init" to initialize MCDR first'.format(core_constant.PACKAGE_NAME))

	# --------------------------
	#         Translate
	# --------------------------

	def get_language(self) -> str:
		return self.translation_manager.language

	def tr(self, translation_key: str, *args, language: Optional[str] = None, allow_failure=True, **kwargs) -> MessageText:
		"""
		Return a translated text corresponded to the translation key and format the text with given args
		If args contains RText element, then the result will be a RText, otherwise the result will be a regular str
		If the translation key is not recognized, the return value will be the translation key itself if allow_failure is True
		:param translation_key: The key of the translation
		:param args: The args to be formatted
		:param language: Specific language to be used in this translation, or the language that MCDR is using will be used
		:param allow_failure: If set to false, a KeyError will be risen if the translation key is not recognized
		:param kwargs: The kwargs to be formatted
		"""
		return self.translation_manager.translate(
			translation_key, args, kwargs,
			allow_failure=allow_failure,
			language=language,
			fallback_language=self.translation_manager.language,
			plugin_translations=self.plugin_manager.registry_storage.translations
		)

	# --------------------------
	#          Loaders
	# --------------------------

	def load_config(self, *, allowed_missing_file: bool = True, echo: bool = True):
		has_missing = self.config.read_config(allowed_missing_file)
		# load the language first to make sure tr() is available
		self.on_config_changed(echo)
		if echo and has_missing:
			for line in self.tr('config.missing_config').splitlines():
				self.logger.warning(line)

	def on_config_changed(self, echo: bool):
		MCColoredFormatter.console_color_disabled = self.config['disable_console_color']
		self.logger.set_debug_options(self.config['debug'])
		if echo and self.config.is_debug_on():
			self.logger.info(self.tr('mcdr_server.on_config_changed.debug_mode_on'))

		self.translation_manager.set_language(self.config['language'])
		if echo:
			self.logger.info(self.tr('mcdr_server.on_config_changed.language_set', self.config['language']))

		self.encoding_method = self.config['encoding'] if self.config['encoding'] is not None else sys.getdefaultencoding()
		self.decoding_method = self.config['decoding'] if self.config['decoding'] is not None else locale.getpreferredencoding()
		if echo:
			self.logger.info(self.tr('mcdr_server.on_config_changed.encoding_decoding_set', self.encoding_method, self.decoding_method))

		self.plugin_manager.set_plugin_directories(self.config['plugin_directories'])
		if echo:
			self.logger.info(self.tr('mcdr_server.on_config_changed.plugin_directories_set', self.encoding_method, self.decoding_method))
			for directory in self.plugin_manager.plugin_directories:
				self.logger.info('- {}'.format(directory))

		self.reactor_manager.register_reactors(self.config['custom_info_reactors'])

		self.server_handler_manager.register_handlers(self.config['custom_handlers'])
		self.server_handler_manager.set_handler(self.config['handler'])
		if echo:
			self.logger.info(self.tr('mcdr_server.on_config_changed.handler_set', self.config['handler']))

		self.connect_rcon()

	def load_plugins(self):
		self.plugin_manager.refresh_all_plugins()
		self.logger.info(self.plugin_manager.last_operation_result.to_rtext(show_path=True))

	def on_plugin_registry_changed(self):
		self.command_manager.clear_command()
		self.plugin_manager.registry_storage.export_commands(self.command_manager.register_command)

	# ---------------------------
	#      General Setters
	# ---------------------------
	# for field read-only access, simply use directly reference
	# but for field writing operation, use setters

	def set_task_executor(self, new_task_executor: TaskExecutor):
		self.task_executor = new_task_executor

	# ---------------------------
	#   State Getters / Setters
	# ---------------------------

	def is_server_running(self):
		return self.server_state.in_state({ServerState.RUNNING, ServerState.STOPPING})

	# Flags

	def is_server_startup(self):
		return MCDReforgedFlag.SERVER_STARTUP in self.flags

	def is_server_rcon_ready(self):
		return MCDReforgedFlag.SERVER_RCON_READY in self.flags

	def is_interrupt(self):
		return MCDReforgedFlag.INTERRUPT in self.flags

	def is_mcdr_exit(self):
		return self.mcdr_in_state(MCDReforgedState.STOPPED)

	def is_mcdr_about_to_exit(self):
		return self.mcdr_in_state({MCDReforgedState.PRE_STOPPED, MCDReforgedState.STOPPED})

	def should_exit_after_stop(self):
		return MCDReforgedFlag.EXIT_AFTER_STOP in self.flags

	def with_flag(self, flag: MCDReforgedFlag):
		self.flags |= flag
		self.logger.debug('Added MCDReforgedFlag {}'.format(flag), option=DebugOption.MCDR)

	def remove_flag(self, flag: MCDReforgedFlag):
		self.flags &= ~flag
		self.logger.debug('Removed MCDReforgedFlag {}'.format(flag), option=DebugOption.MCDR)

	# State

	def server_in_state(self, states):
		return self.server_state.in_state(states)

	def mcdr_in_state(self, states):
		return self.mcdr_state.in_state(states)

	def is_initialized(self):
		return self.mcdr_in_state(MCDReforgedState.INITIALIZED)

	def set_server_state(self, state):
		self.server_state = state
		self.logger.debug('Server state has set to "{}"'.format(state), option=DebugOption.MCDR)

	def set_mcdr_state(self, state):
		self.mcdr_state = state
		self.logger.debug('MCDR state has set to "{}"'.format(state), option=DebugOption.MCDR)

	def should_keep_looping(self):
		"""
		A criterion for sub threads to determine if it should keep looping
		:rtype: bool
		"""
		if self.server_in_state(ServerState.STOPPED):
			if self.is_interrupt():  # if interrupted and stopped
				return False
			if self.should_exit_after_stop():  # natural server stop, or server_interface.exit() by plugin
				return False
			return True
		return not self.is_mcdr_exit()

	# --------------------------
	#      Server Controls
	# --------------------------

	def start_server(self):
		"""
		try to start the server process
		return True if the server process has started successfully
		return False if the server is not able to start

		:return: a bool as above
		:rtype: bool
		"""
		with self.starting_server_lock:
			if self.is_interrupt():
				self.logger.warning(self.tr('mcdr_server.start_server.already_interrupted'))
				return False
			if self.is_server_running():
				self.logger.warning(self.tr('mcdr_server.start_server.start_twice'))
				return False
			if self.is_mcdr_about_to_exit():
				self.logger.warning(self.tr('mcdr_server.start_server.about_to_exit'))
				return False
			cwd = self.config['working_directory']
			if not os.path.isdir(cwd):
				self.logger.error(self.tr('mcdr_server.start_server.cwd_not_existed', cwd))
				return False
			try:
				start_command = self.config['start_command']
				self.logger.info(self.tr('mcdr_server.start_server.starting', start_command))
				self.process = Popen(start_command, cwd=self.config['working_directory'], stdin=PIPE, stdout=PIPE, stderr=STDOUT, shell=True)
			except:
				self.logger.exception(self.tr('mcdr_server.start_server.start_fail'))
				return False
			else:
				self.__on_server_start()
				return True

	def __kill_server(self):
		"""
		Kill the server process group
		"""
		if self.process and self.process.poll() is None:
			self.logger.info(self.tr('mcdr_server.kill_server.killing'))
			try:
				for child in psutil.Process(self.process.pid).children(recursive=True):
					child.kill()
					self.logger.info(self.tr('mcdr_server.kill_server.process_killed', child.pid))
			except psutil.NoSuchProcess:
				pass
			self.process.kill()
			self.logger.info(self.tr('mcdr_server.kill_server.process_killed', self.process.pid))
		else:
			raise IllegalCallError("Server process has already been terminated")

	def interrupt(self):
		"""
		Interrupt MCDR
		The first call will softly stop the server and the later calls will kill the server
		Return if it's the first try
		:rtype: bool
		"""
		self.logger.info('Interrupting, first strike = {}'.format(not self.is_interrupt()))
		self.stop(forced=self.is_interrupt())
		first_try = not self.is_interrupt()
		self.with_flag(MCDReforgedFlag.INTERRUPT)
		return first_try

	def stop(self, forced=False):
		"""
		Stop the server

		:param forced: an optional bool. If it's False (default) MCDR will stop the server by sending the STOP_COMMAND from the
		current handler. If it's True MCDR will just kill the server process group
		"""
		with self.stop_lock:
			if not self.is_server_running():
				self.logger.warning(self.tr('mcdr_server.stop.stop_when_stopped'))
				return
			self.set_server_state(ServerState.STOPPING)
			if not forced:
				try:
					self.send(self.server_handler_manager.get_current_handler().get_stop_command())
				except:
					self.logger.error(self.tr('mcdr_server.stop.stop_fail'))
					forced = True
			if forced:
				try:
					self.__kill_server()
				except IllegalCallError:
					pass

	# --------------------------
	#      Server Logics
	# --------------------------

	def __on_server_start(self):
		self.set_server_state(ServerState.RUNNING)
		self.with_flag(MCDReforgedFlag.EXIT_AFTER_STOP)  # Set after server state is set to RUNNING, or MCDR might have a chance to exit if the server is started by other thread
		self.logger.info(self.tr('mcdr_server.start_server.pid_info', self.process.pid))
		self.reactor_manager.on_server_start()
		self.plugin_manager.dispatch_event(MCDRPluginEvents.SERVER_START, ())

	def __on_server_stop(self):
		return_code = self.process.poll()
		self.logger.info(self.tr('mcdr_server.on_server_stop.show_stopcode', return_code))
		try:
			self.process.stdin.close()
		except Exception as e:
			self.logger.warning('Error when closing stdin: {}'.format(e))
		try:
			self.process.stdout.close()
		except Exception as e:
			self.logger.warning('Error when closing stdout: {}'.format(e))
		self.process = None
		self.set_server_state(ServerState.STOPPED)
		self.remove_flag(MCDReforgedFlag.SERVER_STARTUP | MCDReforgedFlag.SERVER_RCON_READY)  # removes this two
		self.reactor_manager.on_server_stop()
		self.plugin_manager.dispatch_event(MCDRPluginEvents.SERVER_STOP, (return_code,), block=True)

		if self.is_interrupt():
			self.logger.info(self.tr('mcdr_server.on_server_stop.user_interrupted'))
		else:
			self.logger.info(self.tr('mcdr_server.on_server_stop.server_stop'))

	def send(self, text, ending='\n', encoding=None):
		"""
		Send a text to server's stdin if the server is running

		:param text: A str or a bytes you want to send. if text is a str then it will attach the ending parameter to its
		back
		:param str ending: The suffix of a command with a default value \n
		:param str encoding: The encoding method for the text. If it's not given used the method in config
		"""
		if encoding is None:
			encoding = self.encoding_method
		if isinstance(text, str):
			encoded_text = (text + ending).encode(encoding)
		elif isinstance(text, bytes):
			encoded_text = text
		else:
			raise TypeError()
		if self.is_server_running():
			self.process.stdin.write(encoded_text)
			self.process.stdin.flush()
		else:
			self.logger.warning(self.tr('mcdr_server.send.send_when_stopped'))
			self.logger.warning(self.tr('mcdr_server.send.send_when_stopped.text', text if len(text) <= 32 else text[:32] + '...'))

	def __receive(self):
		"""
		Try to receive a str from server's stdout. This will block the current thread
		If server has stopped it will wait up to 10s for the server process to exit, then raise a ServerStopped exception

		:rtype: str
		:raise: ServerStopped
		"""
		try:
			text = next(iter(self.process.stdout))  # type: bytes
		except StopIteration:  # server process has stopped
			for i in range(core_constant.WAIT_TIME_AFTER_SERVER_STDOUT_END_SEC * 10):
				if self.process.poll() is not None:
					break
				time.sleep(0.1)
				if i % 10 == 0:
					self.logger.info(self.tr('mcdr_server.receive.wait_stop'))
			raise ServerStopped()
		else:
			try:
				decoded_text = text.decode(self.decoding_method)  # type: str
			except:
				self.logger.error(self.tr('mcdr_server.receive.decode_fail', text))
				raise
			return decoded_text.rstrip('\n\r').lstrip('\n\r')

	def __tick(self):
		"""
		ticking MCDR:
		try to receive a new line from server's stdout and parse / display / process the text
		"""
		try:
			text = self.__receive()
		except ServerStopped:
			self.__on_server_stop()
			return
		try:
			text = self.server_handler_manager.get_current_handler().pre_parse_server_stdout(text)
		except:
			self.logger.warning(self.tr('mcdr_server.tick.pre_parse_fail'))

		parsed_result: Info
		try:
			parsed_result = self.server_handler_manager.get_current_handler().parse_server_stdout(text)
		except:
			if self.logger.should_log_debug(option=DebugOption.HANDLER):  # traceback.format_exc() is costly
				self.logger.debug('Fail to parse text "{}" from stdout of the server, using raw handler'.format(text), no_check=True)
				for line in traceback.format_exc().splitlines():
					self.logger.debug('    {}'.format(line), no_check=True)
			parsed_result = self.server_handler_manager.get_basic_handler().parse_server_stdout(text)
		else:
			if self.logger.should_log_debug(option=DebugOption.HANDLER):
				self.logger.debug('Parsed text from server stdout:', no_check=True)
				for line in parsed_result.format_text().splitlines():
					self.logger.debug('    {}'.format(line), no_check=True)
		self.server_handler_manager.detect_text(text)
		self.reactor_manager.put_info(parsed_result)

	def __on_mcdr_start(self):
		self.watch_dog.start()
		self.task_executor.start()
		self.preference_manager.load_preferences()
		self.plugin_manager.register_permanent_plugins()
		self.task_executor.execute_on_thread(self.load_plugins, block=True)
		self.plugin_manager.dispatch_event(MCDRPluginEvents.MCDR_START, ())
		if not self.config['disable_console_thread']:
			self.console_handler.start()
		else:
			self.logger.info(self.tr('mcdr_server.on_mcdr_start.console_disabled'))
		if not self.start_server():
			raise ServerStartError()
		self.update_helper.start()
		self.server_handler_manager.start_handler_detection()
		self.set_mcdr_state(MCDReforgedState.RUNNING)

	def __on_mcdr_stop(self):
		try:
			self.set_mcdr_state(MCDReforgedState.PRE_STOPPED)

			self.logger.info(self.tr('mcdr_server.on_mcdr_stop.info'))

			self.plugin_manager.dispatch_event(MCDRPluginEvents.PLUGIN_UNLOADED, ())
			self.task_executor.wait_till_finish_all_task()
			with self.watch_dog.pausing():  # it's ok for plugins to take some time
				self.plugin_manager.dispatch_event(MCDRPluginEvents.MCDR_STOP, (), block=True)

			self.console_handler.stop()
			self.logger.info(self.tr('mcdr_server.on_mcdr_stop.bye'))
		except KeyboardInterrupt:  # I don't know why there sometimes will be a KeyboardInterrupt if MCDR is stopped by ctrl-c
			pass
		except:
			self.logger.exception(self.tr('mcdr_server.on_mcdr_stop.stop_error'))
		finally:
			self.set_mcdr_state(MCDReforgedState.STOPPED)

	def start(self):
		"""
		The entry method to start MCDR
		Try to start the server. if succeeded the console thread will start and MCDR will start ticking

		:raise: IllegalStateError if MCDR is in wrong state
		:raise: ServerStartError if the server is already running or start_server has been called by other
		"""
		if not self.mcdr_in_state(MCDReforgedState.INITIALIZED):
			if self.mcdr_in_state(MCDReforgedState.INITIALIZING):
				raise IllegalStateError('This instance is not fully initialized')
			else:
				raise IllegalStateError('MCDR can only start once')
		self.__main_loop()
		return self.process

	def __main_loop(self):
		"""
		The main loop of MCDR
		"""
		self.__on_mcdr_start()
		while self.should_keep_looping():
			try:
				if self.is_server_running():
					self.__tick()
				else:
					time.sleep(0.01)
			except KeyboardInterrupt:
				self.interrupt()
			except:
				if self.is_interrupt():
					break
				else:
					self.logger.critical(self.tr('mcdr_server.run.error'), exc_info=True)
		self.__on_mcdr_stop()

	def connect_rcon(self):
		self.rcon_manager.disconnect()
		if self.config['rcon']['enable'] and self.is_server_rcon_ready():
			self.rcon_manager.connect(self.config['rcon']['address'], self.config['rcon']['port'], self.config['rcon']['password'])



"""
The basic plain handler
"""
from typing import Any, Optional, Tuple

from mcdreforged.handler.abstract_server_handler import AbstractServerHandler
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation


class BasicHandler(AbstractServerHandler):
	def get_stop_command(self) -> str:
		return ''

	def get_send_message_command(self, target: str, message: Any, server_information: ServerInformation) -> Optional[str]:
		return None

	def get_broadcast_message_command(self, message: Any, server_information: ServerInformation) -> Optional[str]:
		return None

	@classmethod
	def get_content_parsing_formatter(cls):
		raise RuntimeError()

	def parse_server_stdout(self, text: str) -> Info:
		return self._get_server_stdout_raw_result(text)

	def parse_player_joined(self, info):
		return None

	def parse_player_left(self, info):
		return None

	def parse_server_version(self, info: Info) -> Optional[str]:
		return None

	def parse_server_address(self, info: Info) -> Optional[Tuple[str, int]]:
		return None

	def test_server_startup_done(self, info) -> bool:
		return False

	def test_rcon_started(self, info: Info) -> bool:
		return False

	def test_server_stopping(self, info: Info) -> bool:
		return False



import collections
from enum import Enum
from typing import Optional, Dict, List, Union


PermissionParam = Union[str, int]


class PermissionLevelItem:
	def __init__(self, name: str, level: int):
		self.name = name
		self.level = level

	def __str__(self):
		return '{} ({})'.format(self.level, self.name)

	def __repr__(self):
		return 'Permission[name={},level={}]'.format(self.name, self.level)

	def __lt__(self, other):
		if not isinstance(other, type(self)):
			return False
		return self.level < other.level


class PermissionLevel:
	class __Storage(Enum):
		GUEST_	= PermissionLevelItem('guest', 0)
		USER_	= PermissionLevelItem('user', 1)
		HELPER_	= PermissionLevelItem('helper', 2)
		ADMIN_	= PermissionLevelItem('admin', 3)
		OWNER_	= PermissionLevelItem('owner', 4)

	GUEST = __Storage.GUEST_.value.level
	USER = __Storage.USER_.value.level
	HELPER = __Storage.HELPER_.value.level
	ADMIN = __Storage.ADMIN_.value.level
	OWNER = __Storage.OWNER_.value.level

	INSTANCES = [item.value for item in __Storage]  # type: List[PermissionLevelItem]
	LEVELS = [inst.level for inst in INSTANCES]  # type: List[int]
	NAMES = [inst.name for inst in INSTANCES]  # type: List[str]
	__NAME_DICT = collections.OrderedDict(zip(NAMES, INSTANCES))  # type: Dict[str, PermissionLevelItem]
	__LEVEL_DICT = collections.OrderedDict(zip(LEVELS, INSTANCES))  # type: Dict[int, PermissionLevelItem]

	MAXIMUM_LEVEL = LEVELS[-1]  # type: int
	MINIMUM_LEVEL = LEVELS[0]  # type: int
	MCDR_CONTROL_LEVEL = ADMIN  # type: int
	PHYSICAL_SERVER_CONTROL_LEVEL = OWNER  # type: int
	CONSOLE_LEVEL = MAXIMUM_LEVEL  # type: int
	PLUGIN_LEVEL = MAXIMUM_LEVEL  # type: int

	@classmethod
	def __check_range(cls, level: int):
		if cls.MINIMUM_LEVEL <= level <= cls.MAXIMUM_LEVEL:
			pass
		else:
			raise ValueError('Value {} out of range [{}, {}]'.format(level, cls.MINIMUM_LEVEL, cls.MAXIMUM_LEVEL))

	@classmethod
	def from_value(cls, value: PermissionParam):
		"""
		Convert any type of permission level into int value. Examples:
			'guest'	-> 0
			'admin'	-> 3
			'1'		-> 1
			2		-> 2
		If the argument is invalid return None

		:param value: a permission related object
		:type value: str or int
		:rtype: PermissionLevelItem
		"""
		level = None
		if isinstance(value, str):
			if value.isdigit():
				value = int(value)
			elif value in cls.NAMES:
				level = cls.__NAME_DICT[value]
		if isinstance(value, int):
			cls.__check_range(value)
			level = cls.__LEVEL_DICT[value]
		if level is None:
			raise TypeError('Unsupported value for {}: {}'.format(cls.__name__, value))
		return level

	@classmethod
	def get_level(cls, value: PermissionParam) -> Optional[PermissionLevelItem]:
		"""
		Fail-proof version of from_value
		"""
		try:
			return cls.from_value(value)
		except (TypeError, ValueError):
			return None



from mcdreforged.plugin.plugin_event import LiteralEvent, PluginEvent, MCDRPluginEvents

__all__ = [
	# If you want to customize, inherit this
	'PluginEvent',

	# A simple event type
	'LiteralEvent',

	# Built-in MCDR events
	'MCDRPluginEvents',
]



from mcdreforged.command.builder import command_builder_util
from mcdreforged.command.builder.exception import LiteralNotMatch, NumberOutOfRange, IllegalArgument, EmptyText, \
	UnknownCommand, UnknownArgument, CommandSyntaxError, UnknownRootArgument, RequirementNotMet, IllegalNodeOperation, \
	CommandError, InvalidNumber, InvalidInteger, InvalidFloat, UnclosedQuotedString, IllegalEscapesUsage, InvalidBoolean
from mcdreforged.command.builder.nodes.arguments import Number, Integer, Float, Text, QuotableText, \
	GreedyText, Boolean
from mcdreforged.command.builder.nodes.basic import AbstractNode, Literal, ParseResult, CommandContext, ArgumentNode

__all__ = [
	# ------------------
	#   Argument Nodes
	# ------------------

	'AbstractNode', 'ArgumentNode',
	'Literal',
	'Number', 'Integer', 'Float',
	'Text', 'QuotableText', 'GreedyText',
	'Boolean',

	# ------------------
	#     Exceptions
	# ------------------

	'IllegalNodeOperation',

	'CommandError',
	'UnknownCommand', 'UnknownArgument', 'UnknownRootArgument', 'RequirementNotMet',

	'CommandSyntaxError',
	'IllegalArgument', 'LiteralNotMatch',
	'NumberOutOfRange', 'InvalidNumber', 'InvalidInteger', 'InvalidFloat',
	'IllegalEscapesUsage', 'UnclosedQuotedString', 'EmptyText',
	'InvalidBoolean',

	# ------------------
	#       Utils
	# ------------------

	'CommandContext',
	'command_builder_util',
	'ParseResult'
]



import os
import sys
import zipimport
from io import BytesIO
from typing import IO, Collection, TYPE_CHECKING, Optional
from zipfile import ZipFile

from mcdreforged.plugin.type.multi_file_plugin import MultiFilePlugin
from mcdreforged.utils.exception import IllegalPluginStructure

if TYPE_CHECKING:
	from mcdreforged.plugin.plugin_manager import PluginManager


class PackedPlugin(MultiFilePlugin):
	def __init__(self, plugin_manager: 'PluginManager', file_path: str):
		super().__init__(plugin_manager, file_path)
		self.__zip_file_cache = None  # type: Optional[ZipFile]

	@property
	def __zip_file(self) -> ZipFile:
		if self.__zip_file_cache is None:
			with open(self.plugin_path, 'rb') as file_handler:
				content = BytesIO()
				content.write(file_handler.read())
			self.__zip_file_cache = ZipFile(content)
		return self.__zip_file_cache

	@classmethod
	def __format_path(cls, path: str):
		return path.replace('\\', '/')

	def _reset(self):
		super()._reset()
		if self.__zip_file_cache is not None:
			self.__zip_file_cache.close()
		self.__zip_file_cache = None

	def open_file(self, file_path: str) -> IO[bytes]:
		return self.__zip_file.open(self.__format_path(file_path), 'r')

	def list_directory(self, directory_name: str) -> Collection[str]:
		result = []
		directory_name = self.__format_path(directory_name).rstrip('/\\') + '/'
		for file_info in self.__zip_file_cache.infolist():
			# is inside the dir and is directly inside
			if file_info.filename.startswith(directory_name):
				file_name = file_info.filename.replace(directory_name, '', 1)
				if len(file_name) > 0 and '/' not in file_name.rstrip('/'):
					result.append(file_name)
		return result

	def _check_subdir_legality(self):
		for file_info in self.__zip_file_cache.infolist():
			if file_info.is_dir():
				package_name: str = file_info.filename[:-1]  # removing the ending '/'
				if '/' in package_name:  # not at root
					continue
				try:
					init_info = self.__zip_file_cache.getinfo(os.path.join(package_name, '__init__.py'))
				except KeyError:
					is_module = False
				else:
					is_module = not init_info.is_dir()
				if is_module and package_name != self.get_id():
					raise IllegalPluginStructure('Packed plugin cannot contain other package: found package {}'.format(package_name))

	# noinspection PyProtectedMember,PyUnresolvedReferences
	def _on_unload(self):
		super()._on_unload()
		try:
			for path in list(sys.path_importer_cache.keys()):
				if path.startswith(self.plugin_path):
					sys.path_importer_cache.pop(path)
			if self.plugin_path in zipimport._zip_directory_cache:
				zipimport._zip_directory_cache.pop(self.plugin_path)
		except KeyError:
			self.mcdr_server.logger.exception('Fail to clean zip import cache for {}'.format(self))



import json
import re


def remove_suffix(text: str, suffix: str):
	pos = text.rfind(suffix)
	return text[:pos] if pos >= 0 else text


def clean_minecraft_color_code(text):
	return re.sub('§[a-z0-9]', '', str(text))


def clean_console_color_code(text):
	return re.sub(r'\033\[(\d+(;\d+)?)?m', '', text)


def hump_to_underline(name: str) -> str:
	"""
	ThisIsAnHumpName -> this_is_an_hump_name
	"""
	return re.sub(r'([a-z]|\d)([A-Z])', r'\1_\2', name).lower()


def auto_quotes(text: str):
	if not isinstance(text, str):
		raise TypeError('Parameter text should be a str, but {} founded'.format(type(text)))
	if ' ' in text:
		return json.dumps(text)
	else:
		return text



"""
MCDR config file stuffs
"""
from logging import Logger
from typing import Any

from mcdreforged.utils.yaml_data_storage import YamlDataStorage

CONFIG_FILE = 'config.yml'
DEFAULT_CONFIG_RESOURCE_PATH = 'resources/default_config.yml'


class Config(YamlDataStorage):
	def __init__(self, logger: Logger):
		super().__init__(logger, CONFIG_FILE, DEFAULT_CONFIG_RESOURCE_PATH)

	def read_config(self, allowed_missing_file: bool):
		return self._load_data(allowed_missing_file)

	def __getitem__(self, option: str):
		return self._data[option]

	def set_value(self, option: str, value: Any):
		if option in self._data and type(self[option]) == type(value):
			self._data[option] = value
		elif option not in self._data:
			raise KeyError('Cannot set option {} since the config does not contains it'.format(option))
		else:
			raise ValueError('Cannot set option {} in type {} to value {} in type {}'.format(option, type(self[option]), value, type(value)))

	# -------------------------
	#   Actual data analyzers
	# -------------------------

	def is_debug_on(self):
		for value in self._data['debug']:
			if value is True:
				return True
		return False



from abc import ABC
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Optional

from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.translation.translation_text import RTextMCDRTranslation
from mcdreforged.utils import misc_util

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer
	from mcdreforged.info_reactor.info import Info
	from mcdreforged.plugin.server_interface import ServerInterface
	from mcdreforged.plugin.type.plugin import AbstractPlugin
	from mcdreforged.preference.preference_manager import PreferenceItem


class CommandSource(ABC):
	@property
	def is_player(self) -> bool:
		raise NotImplementedError()

	@property
	def is_console(self) -> bool:
		raise NotImplementedError()

	def get_server(self) -> 'ServerInterface':
		raise NotImplementedError()

	def get_permission_level(self) -> int:
		raise NotImplementedError()

	def get_preference(self) -> 'PreferenceItem':
		return self.get_server().get_preference(self)

	@contextmanager
	def preferred_language_context(self):
		with RTextMCDRTranslation.language_context(self.get_preference().language):
			yield

	def has_permission(self, level: int) -> bool:
		return self.get_permission_level() >= level

	def has_permission_higher_than(self, level: int) -> bool:
		return self.get_permission_level() > level

	def reply(self, message: Any, **kwargs) -> None:
		"""
		Reply to the command source
		:param message: The message you want to send. It will be mapped with str() unless it's a RTextBase
		"""
		raise NotImplementedError()


class InfoCommandSource(CommandSource, ABC):
	"""
	Command source generated from info
	"""
	def __init__(self, mcdr_server: 'MCDReforgedServer', info: 'Info'):
		self._mcdr_server = mcdr_server
		self.__info = info

	def get_info(self) -> 'Info':
		return self.__info

	def get_server(self) -> 'ServerInterface':
		return self._mcdr_server.basic_server_interface

	def get_permission_level(self) -> int:
		return self._mcdr_server.permission_manager.get_permission(self)

	def __str__(self):
		raise NotImplementedError()

	def __repr__(self):
		raise NotImplementedError()


class PlayerCommandSource(InfoCommandSource):
	def __init__(self, mcdr_server, info, player: str):
		if not info.is_from_server:
			raise TypeError('{} should be built from server info'.format(self.__class__.__name__))
		super().__init__(mcdr_server, info)
		self.player = player

	@property
	def is_player(self) -> bool:
		return True

	@property
	def is_console(self) -> bool:
		return False

	def reply(self, message: Any, *, encoding: Optional[str] = None, **kwargs):
		"""
		:keyword encoding: encoding method for server_interface.tell
		"""
		self._mcdr_server.basic_server_interface.tell(self.player, message, encoding=encoding)

	def __str__(self):
		return 'Player {}'.format(self.player)

	def __repr__(self):
		return '{}[player={},info_id={}]'.format(type(self).__name__, self.player, self.get_info().id)


class ConsoleCommandSource(InfoCommandSource):
	def __init__(self, mcdr_server, info):
		if not info.is_from_console:
			raise TypeError('{} should be built from console info'.format(self.__class__.__name__))
		super().__init__(mcdr_server, info)

	@property
	def is_player(self) -> bool:
		return False

	@property
	def is_console(self) -> bool:
		return True

	def reply(self, message: Any, *, console_text: Optional[Any] = None, **kwargs):
		"""
		:keyword console_text: If it's specified, use it instead of param message
		"""
		if console_text is not None:
			message = console_text
		with self.preferred_language_context():
			misc_util.print_text_to_console(self._mcdr_server.logger, message)

	def __str__(self):
		return 'Console'

	def __repr__(self):
		return '{}[info_id={}]'.format(type(self).__name__, self.get_info().id)


class PluginCommandSource(CommandSource):
	def __init__(self, server: 'ServerInterface', plugin: Optional['AbstractPlugin'] = None):
		self.__server = server.as_basic_server_interface()
		self.__logger = self.__server.logger
		self.__plugin = plugin

	@property
	def is_player(self) -> bool:
		return False

	@property
	def is_console(self) -> bool:
		return False

	def get_server(self) -> 'ServerInterface':
		return self.__server

	def get_permission_level(self) -> int:
		return PermissionLevel.PLUGIN_LEVEL

	def reply(self, message: Any, **kwargs) -> None:
		misc_util.print_text_to_console(self.__logger, message)

	def __str__(self):
		return 'Plugin' if self.__plugin is None else 'Plugin {}'.format(self.__plugin)

	def __repr__(self):
		return '{}[plugin={}]'.format(type(self).__name__, self.__plugin)



from mcdreforged.handler.impl.abstract_minecraft_handler import AbstractMinecraftHandler


class VanillaHandler(AbstractMinecraftHandler):
	@classmethod
	def get_content_parsing_formatter(cls):
		return '[{hour:d}:{min:d}:{sec:d}] [{thread}/{logging}]: {content}'



import re
from typing import Any, Optional

from parse import parse

from mcdreforged.handler.abstract_server_handler import AbstractServerHandler
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation


class BungeecordHandler(AbstractServerHandler):
	def get_stop_command(self) -> str:
		return 'end'

	def get_send_message_command(self, target: str, message: Any, server_information: ServerInformation) -> Optional[str]:
		return None

	def get_broadcast_message_command(self, message: Any, server_information: ServerInformation) -> Optional[str]:
		from mcdreforged.handler.impl import VanillaHandler
		return 'alertraw {}'.format(VanillaHandler.format_message(message))

	@classmethod
	def get_content_parsing_formatter(cls):
		# 09:00:02 [信息] Listening on /0.0.0.0:25565
		# 09:00:01 [信息] [Steve] -> UpstreamBridge has disconnected
		return '{hour:d}:{min:d}:{sec:d} [{logging}] {content}'

	def pre_parse_server_stdout(self, text):
		text = super().pre_parse_server_stdout(text)
		match = re.match(r'>*\r', text)
		if match is not None:
			text = text.replace(match.group(), '', 1)
		return text

	def parse_player_joined(self, info: Info) -> Optional[str]:
		# [Steve,/127.0.0.1:3631] <-> InitialHandler has connected
		if not info.is_user:
			parsed = parse('[{name},/{ip}] <-> InitialHandler has connected', info.content)
			if parsed is not None:
				return parsed['name']
		return None

	def parse_player_left(self, info):
		# [Steve] -> UpstreamBridge has disconnected
		if not info.is_user:
			parsed = parse('[{name}] -> UpstreamBridge has disconnected', info.content)
			if parsed is not None:
				return parsed['name']
		return None

	def parse_server_version(self, info: Info):
		return None

	def parse_server_address(self, info: Info):
		# Listening on /0.0.0.0:25577
		if not info.is_user:
			parsed = parse('Listening on /{}:{:d}', info.content)
			if parsed is not None:
				return parsed[0], parsed[1]
		return None

	def test_server_startup_done(self, info: Info) -> bool:
		# Listening on /0.0.0.0:25577
		return not info.is_user and re.fullmatch(r'Listening on /[0-9.]+:[0-9]+', info.content) is not None

	def test_rcon_started(self, info: Info) -> bool:
		return self.test_server_startup_done(info)

	def test_server_stopping(self, info: Info) -> bool:
		# Closing listener [id: 0x3acae0b0, L:/0:0:0:0:0:0:0:0:25565]
		return not info.is_user and re.fullmatch(r'Closing listener \[id: .+, L:[\d:/]+]', info.content) is not None



import sys
from threading import RLock, Lock
from typing import TYPE_CHECKING, Optional, Iterable, Any, List

from prompt_toolkit import PromptSession
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import CompletionState, Buffer
from prompt_toolkit.completion import Completion, CompleteEvent, WordCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.layout import Dimension
# noinspection PyProtectedMember
from prompt_toolkit.layout.menus import MultiColumnCompletionMenuControl
from prompt_toolkit.layout.processors import Processor, TransformationInput, Transformation
from prompt_toolkit.output.vt100 import Vt100_Output
from prompt_toolkit.patch_stdout import StdoutProxy
from prompt_toolkit.shortcuts import CompleteStyle

from mcdreforged.command.builder.nodes.basic import CommandSuggestions
from mcdreforged.command.command_source import CommandSource
from mcdreforged.executor.thread_executor import ThreadExecutor
from mcdreforged.info_reactor.info import Info
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.plugin.server_interface import ServerInterface
from mcdreforged.utils import misc_util
from mcdreforged.utils.logger import DebugOption, SyncStdoutStreamHandler

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer
	from mcdreforged.command.command_manager import CommandManager


class ConsoleHandler(ThreadExecutor):
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		super().__init__(mcdr_server)
		self.console_kit = PromptToolkitWrapper(self)

	def loop(self):
		if self.mcdr_server.config['advanced_console']:
			self.console_kit.start_kits()
		super().loop()

	def stop(self):
		super().stop()
		self.console_kit.stop_kits()

	def tick(self):
		try:
			inputs = self.console_kit.get_input()
			if not isinstance(inputs, list):
				raise IOError()
			for text in inputs:
				if not isinstance(text, str):
					raise IOError()
				try:
					parsed_result: Info = self.mcdr_server.server_handler_manager.get_current_handler().parse_console_command(text)
				except:
					self.mcdr_server.logger.exception(self.mcdr_server.tr('console_handler.parse_fail', text))
				else:
					if self.mcdr_server.logger.should_log_debug(DebugOption.HANDLER):
						self.mcdr_server.logger.debug('Parsed text from {}:'.format(type(self).__name__), no_check=True)
						for line in parsed_result.format_text().splitlines():
							self.mcdr_server.logger.debug('	{}'.format(line), no_check=True)
					self.mcdr_server.reactor_manager.put_info(parsed_result)
		except (KeyboardInterrupt, EOFError, SystemExit, IOError) as error:
			if self.mcdr_server.is_server_running():
				self.mcdr_server.logger.critical('Critical exception caught in {}: {} {}'.format(type(self).__name__, type(error).__name__, error))
				if not self.mcdr_server.interrupt():  # not first try
					self.mcdr_server.logger.error('Console thread stopped')
					self.stop()
		except:
			self.mcdr_server.logger.exception(self.mcdr_server.tr('console_handler.error'))


class ConsoleSuggestionCommandSource(CommandSource):
	@property
	def is_player(self) -> bool:
		return False

	@property
	def is_console(self) -> bool:
		return True

	def get_server(self) -> 'ServerInterface':
		return ServerInterface.get_instance()

	def get_permission_level(self) -> int:
		return PermissionLevel.CONSOLE_LEVEL

	def reply(self, message: Any, **kwargs) -> None:
		pass


class CachedSuggestionProvider:
	def __init__(self, command_manager: 'CommandManager'):
		self.__command_manager = command_manager
		self.__lock = Lock()
		self.__calc_lock = Lock()
		self.__cache_input: Optional[str] = None
		self.__cache_suggestion: Optional[CommandSuggestions] = None

	def suggest(self, input_: str) -> CommandSuggestions:
		with self.__lock:
			if input_ == self.__cache_input:
				return self.__cache_suggestion
		acq = self.__calc_lock.acquire(blocking=False)
		if not acq:
			return CommandSuggestions()  # empty suggestion
		try:
			suggestion = self.__command_manager.suggest_command(input_, ConsoleSuggestionCommandSource())
			with self.__lock:
				self.__cache_input = input_
				self.__cache_suggestion = suggestion
			return self.__cache_suggestion
		finally:
			self.__calc_lock.release()


class CommandCompleter(WordCompleter):
	def __init__(self, suggester: CachedSuggestionProvider):
		super().__init__([], sentence=True)
		self.suggester = suggester

	def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
		input_ = document.current_line_before_cursor
		suggestions = self.suggester.suggest(input_)
		self.words = sorted(misc_util.unique_list([suggestion.command for suggestion in suggestions]))
		self.display_dict = dict([(suggestion.command, suggestion.suggest_input) for suggestion in suggestions])
		return super().get_completions(document, complete_event)


class CommandArgumentSuggester(Processor):
	def __init__(self, suggester: CachedSuggestionProvider):
		super().__init__()
		self.suggester = suggester

	def apply_transformation(self, ti: TransformationInput) -> Transformation:
		# last line and cursor at end
		if ti.lineno == ti.document.line_count - 1 and ti.document.is_cursor_at_the_end:
			buffer = ti.buffer_control.buffer
			if not buffer.suggestion:
				input_ = ti.document.text
				suggestions = self.suggester.suggest(input_)
				if suggestions.complete_hint is not None:
					return Transformation(fragments=ti.fragments + [('class:auto-suggestion', suggestions.complete_hint)])
		return Transformation(fragments=ti.fragments)


class MCDRPromptSession(PromptSession):
	def __init__(self, console_handler: ConsoleHandler):
		self.__console_handler: ConsoleHandler = console_handler
		# noinspection PyArgumentEqualDefault
		super().__init__(
			complete_in_thread=True,  # this will do the ThreadedCompleter wrapping automatically
			complete_while_typing=True,
			complete_style=CompleteStyle.MULTI_COLUMN,
			reserve_space_for_menu=3
		)
		suggester = CachedSuggestionProvider(self.__console_handler.mcdr_server.command_manager)
		self.completer = CommandCompleter(suggester)
		self.input_processors = [
			CommandArgumentSuggester(suggester),
			self.__get_complete_state_checker()
		]
		self.has_complete_this_line = False

	@staticmethod
	def _has_completion_now(buffer: Buffer) -> bool:
		# Completion calculation might be happening right now in another thread, so make a reference copy first.
		complete_state: Optional[CompletionState] = buffer.complete_state
		# Similarly, in addition to checking if it's None, we should checking its completions field as well
		return complete_state is not None and complete_state.completions

	def _get_default_buffer_control_height(self) -> Dimension:
		dim = super()._get_default_buffer_control_height()
		# When there's no complete this line, don't reverse space for the autocompletion menu
		if not self.has_complete_this_line and not self._has_completion_now(self.default_buffer):
			dim = Dimension()
		return dim

	def __get_complete_state_checker(self) -> Processor:
		class __CompleteStateChecker(Processor):
			def apply_transformation(self, ti: TransformationInput) -> Transformation:
				if len(ti.document.text) == 0:
					session.has_complete_this_line = False
				if session._has_completion_now(buff):
					session.has_complete_this_line = True
				return Transformation(fragments=ti.fragments)

		buff = self.default_buffer
		session = self
		return __CompleteStateChecker()


class PromptToolkitWrapper:
	def __init__(self, console_handler: ConsoleHandler):
		self.__console_handler: ConsoleHandler = console_handler
		self.__tr = console_handler.mcdr_server.tr
		self.__logger = console_handler.mcdr_server.logger
		self.pt_enabled = False
		self.stdout_proxy: Optional[StdoutProxy] = None
		self.prompt_session: Optional[MCDRPromptSession] = None
		self.__real_stdout = None
		self.__real_stderr = None
		self.__promoting = RLock()  # more for a status check

	def start_kits(self):
		try:
			self.__tweak_kits()
			self.prompt_session = MCDRPromptSession(self.__console_handler)
			self.stdout_proxy = StdoutProxy(sleep_between_writes=0.01, raw=True)
		except:
			self.__logger.exception('Failed to enable advanced console, switch back to basic input')
		else:
			self.__real_stdout = sys.stdout
			self.__real_stderr = sys.stderr
			sys.stdout = self.stdout_proxy
			sys.stderr = self.stdout_proxy
			SyncStdoutStreamHandler.update_stdout()
			self.pt_enabled = True
			self.__logger.debug('Prompt Toolkits enabled')

	@staticmethod
	def __tweak_kits():
		# monkey patch to yeet the bell sound
		Vt100_Output.bell = lambda self_: None

		# monkey patch to set min_rows to 2
		# since with reserve_space_for_menu=3, 2 row of completion menu can be displayed
		def min_rows_is_2(*args, **kwargs):
			kwargs['min_rows'] = 2
			real_ctor(*args, **kwargs)

		# noinspection PyTypeChecker
		real_ctor = MultiColumnCompletionMenuControl.__init__
		MultiColumnCompletionMenuControl.__init__ = min_rows_is_2

	def stop_kits(self):
		if self.pt_enabled:
			self.pt_enabled = False
			self.__logger.info(self.__tr('console_handler.stopping_kits'))
			self.stdout_proxy.close()
			sys.stdout = self.__real_stdout
			sys.stderr = self.__real_stderr
			SyncStdoutStreamHandler.update_stdout()
			pt_app = get_app()
			if pt_app.is_running:
				try:
					pt_app.exit()
				except:
					self.__logger.exception('Fail to stop prompt toolkit app')
			with self.__promoting:
				# make sure in the console thread, prompt_session.prompt() ends
				pass

	def get_input(self) -> List[str]:
		if self.pt_enabled:
			assert self.__console_handler.is_on_thread()
			with self.__promoting:
				# It's possible for prompt-toolkit to return a string wth multiple lines when pasting (#146)
				input_ = self.prompt_session.prompt('> ')
				if input_ is None:
					input_ = ''
				return input_.splitlines()
		else:
			return [input()]



import copy
from abc import ABC
from enum import EnumMeta
from threading import Lock
from typing import Union, TypeVar, List, Dict, Type, get_type_hints, Any

T = TypeVar('T')


def _get_type_hints(cls: Type):
	try:
		return get_type_hints(cls)
	except:
		return get_type_hints(cls, globalns={})


def _get_origin(cls: Type):
	return getattr(cls, '__origin__', None)


def _get_args(cls: Type) -> tuple:
	return getattr(cls, '__args__', ())


def serialize(obj) -> Union[None, int, float, str, list, dict]:
	if type(obj) in (type(None), int, float, str, bool):
		return obj
	elif isinstance(obj, list) or isinstance(obj, tuple):
		return list(map(serialize, obj))
	elif isinstance(obj, dict):
		return dict(map(lambda t: (t[0], serialize(t[1])), obj.items()))
	elif isinstance(obj.__class__, EnumMeta):
		return obj.name
	try:
		attr_dict = vars(obj).copy()
		# don't serialize protected fields
		for attr_name in list(attr_dict.keys()):
			if attr_name.startswith('_'):
				attr_dict.pop(attr_name)
	except:
		raise TypeError('Unsupported input type {}'.format(type(obj))) from None
	else:
		return serialize(attr_dict)


_BASIC_CLASSES = (type(None), bool, int, float, str, list, dict)


def deserialize(data, cls: Type[T], *, error_at_missing=False, error_at_redundancy=False) -> T:
	# in case None instead of NoneType is passed
	if cls is None:
		cls = type(None)
	# if its type is Any, then simply return the data
	if cls is Any:
		return data
	# Union
	# Unpack Union first since the target class is not confirmed yet
	elif _get_origin(cls) == Union:
		for possible_cls in _get_args(cls):
			try:
				return deserialize(data, possible_cls, error_at_missing=error_at_missing, error_at_redundancy=error_at_redundancy)
			except (TypeError, ValueError):
				pass
		raise TypeError('Data in type {} cannot match any candidate of target class {}'.format(type(data), cls))
	# Element (None, int, float, str, list, dict)
	# For list and dict, since it doesn't have any type hint, we choose to simply return the data
	elif cls in _BASIC_CLASSES and type(data) is cls:
		return data
	# float thing
	elif cls is float and isinstance(data, int):
		return float(data)
	# List
	elif _get_origin(cls) == List[int].__origin__ and isinstance(data, list):
		element_type = _get_args(cls)[0]
		return list(map(lambda e: deserialize(e, element_type, error_at_missing=error_at_missing, error_at_redundancy=error_at_redundancy), data))
	# Dict
	elif _get_origin(cls) == Dict[int, int].__origin__ and isinstance(data, dict):
		key_type = _get_args(cls)[0]
		val_type = _get_args(cls)[1]
		instance = {}
		for key, value in data.items():
			deserialized_key = deserialize(key, key_type, error_at_missing=error_at_missing, error_at_redundancy=error_at_redundancy)
			deserialized_value = deserialize(value, val_type, error_at_missing=error_at_missing, error_at_redundancy=error_at_redundancy)
			instance[deserialized_key] = deserialized_value
		return instance
	# Enum
	elif isinstance(cls, EnumMeta) and isinstance(data, str):
		return cls[data]
	# Object
	elif cls not in _BASIC_CLASSES and isinstance(cls, type) and isinstance(data, dict):
		try:
			result = cls()
		except:
			raise TypeError('Failed to construct instance of class {}'.format(type(cls)))
		input_key_set = set(data.keys())
		for attr_name, attr_type in _get_type_hints(cls).items():
			if not attr_name.startswith('_'):
				if attr_name in data:
					result.__setattr__(attr_name, deserialize(data[attr_name], attr_type, error_at_missing=error_at_missing, error_at_redundancy=error_at_redundancy))
					input_key_set.remove(attr_name)
				elif error_at_missing:
					raise ValueError('Missing attribute {} for class {} in input object {}'.format(attr_name, cls, data))
				elif hasattr(cls, attr_name):
					result.__setattr__(attr_name, copy.copy(getattr(cls, attr_name)))
		if error_at_redundancy and len(input_key_set) > 0:
			raise ValueError('Redundancy attributes {} for class {} in input object {}'.format(input_key_set, cls, data))
		if isinstance(result, Serializable):
			result.on_deserialization()
		return result
	else:
		raise TypeError('Unsupported input type: expected class {} but found data with class {}'.format(cls, type(data)))


class Serializable(ABC):
	__annotations_cache: dict = None
	__annotations_lock = Lock()

	def __init__(self, **kwargs):
		for key in kwargs.keys():
			if key not in self.get_annotations_fields():
				raise KeyError('Unknown key received in __init__ of class {}: {}'.format(self.__class__, key))
		vars(self).update(kwargs)

	@classmethod
	def __get_annotation_dict(cls) -> dict:
		public_fields = {}
		for attr_name, attr_type in _get_type_hints(cls).items():
			if not attr_name.startswith('_'):
				public_fields[attr_name] = attr_type
		return public_fields

	@classmethod
	def get_annotations_fields(cls) -> Dict[str, Type]:
		with cls.__annotations_lock:
			if cls.__annotations_cache is None:
				cls.__annotations_cache = cls.__get_annotation_dict()
		return cls.__annotations_cache

	def serialize(self) -> dict:
		return serialize(self)

	@classmethod
	def deserialize(cls, data: dict, **kwargs):
		return deserialize(data, cls, **kwargs)

	def update_from(self, data: dict):
		vars(self).update(vars(self.deserialize(data)))

	@classmethod
	def get_default(cls):
		return cls.deserialize({})

	def on_deserialization(self):
		"""
		Invoked after being deserialized
		"""
		pass



import os
from typing import IO

from mcdreforged.plugin.type.multi_file_plugin import MultiFilePlugin
from mcdreforged.utils.exception import IllegalPluginStructure


class DirectoryPlugin(MultiFilePlugin):
	def open_file(self, file_path: str) -> IO[bytes]:
		return open(os.path.join(self.plugin_path, file_path), 'rb')

	def _check_subdir_legality(self):
		for package_name in os.listdir(self.plugin_path):
			path = os.path.join(self.plugin_path, package_name)
			if os.path.isdir(path):
				is_module = os.path.isfile(os.path.join(path, '__init__.py'))
				if is_module and package_name != self.get_id():
					raise IllegalPluginStructure('Packed plugin cannot contain other package: found package {}'.format(package_name))

	def plugin_exists(self):
		return os.path.isdir(self.plugin_path)

	def file_changed(self):
		# It's not that easy to access all files in the directory
		# So just consider Directory Plugin to be always changed
		return True

	def calculate_file_modify_time(self):
		return None



import re
from typing import Optional, Any

from parse import parse

from mcdreforged.handler.impl.vanilla_handler import VanillaHandler
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.minecraft.rtext import RTextBase
from mcdreforged.utils import string_util


class Beta18Handler(VanillaHandler):
	@classmethod
	def format_message(cls, message: Any) -> str:
		if isinstance(message, RTextBase):
			content = message.to_plain_text()
		else:
			content = str(message)
		return string_util.clean_minecraft_color_code(content)

	def get_send_message_command(self, target: str, message: Any, server_information: ServerInformation) -> Optional[str]:
		return 'tell {} {}'.format(target, self.format_message(message))

	def get_broadcast_message_command(self, message: Any, server_information: ServerInformation) -> Optional[str]:
		return 'say {}'.format(self.format_message(message))

	# 2020-12-27 02:56:35 [INFO] Loading properties
	# 2020-12-27 02:57:42 [INFO] <Steve> ping
	@classmethod
	def get_content_parsing_formatter(cls):
		return '{y:d}-{m:d}-{d:d} {hour:d}:{min:d}:{sec:d} [{logging}] {content}'

	def parse_player_joined(self, info):
		# Steve [/127.0.0.1:2993] logged in with entity id 3827 at (-130.5, 69.0, 253.5)
		if not info.is_user:
			# there's an extra space character after {name}
			parsed = parse('{name} [{}] logged in with entity id {} at ({})', info.content)
			if parsed is not None and self._verify_player_name(parsed['name']):
				return parsed['name']
		return None

	def parse_player_left(self, info: Info):
		# Steve lost connection: disconnect.quitting
		if info.is_from_server and parse('{name} lost connection: {}', info.content) is not None:
			return info.content.split(' ')[0]
		return None

	def test_server_startup_done(self, info: Info):
		# Done (6368115300ns)! For help, type "help" or "?"
		if info.is_user:
			return False
		match = re.fullmatch(r'Done \([0-9.]*ns\)! For help, type "help" or "\?"', info.content)
		return match is not None

	def test_rcon_started(self, info: Info):
		return False

	def test_server_stopping(self, info: Info):
		# Stopping server
		return not info.is_user and info.content == 'Stopping server'



from typing import TypeVar, Callable

T = TypeVar('T')


class LazyItem:
	__NONE = object()

	def __init__(self, provider: Callable[[], T]):
		self.__provider = provider
		self.__item = self.__NONE

	def get(self) -> T:
		if self.__item is self.__NONE:
			self.__item = self.__provider()
		return self.__item



"""
Plugin Version
"""
import re
from typing import List, Callable, Tuple, Optional


# beta.3 -> (beta, 3), random -> (random, None)
class ExtraElement:
	DIVIDER = '.'
	body: str
	num: Optional[int]

	def __init__(self, segment_str: str):
		segments = segment_str.rsplit(self.DIVIDER, 1)
		try:
			self.body, self.num = segments[0], int(segments[1])
		except (IndexError, ValueError):
			self.body, self.num = segment_str, None

	def __str__(self):
		if self.num is None:
			return self.body
		return '{}{}{}'.format(self.body, self.DIVIDER, self.num)

	def __lt__(self, other):
		if not isinstance(other, type(self)):
			raise TypeError()
		if self.num is None or other.num is None:
			return str(self) < str(other)
		else:
			return (self.body, self.num) < (other.body, other.num)


class Version:
	EXTRA_ID_PATTERN = re.compile(r'|[-+0-9A-Za-z]+(\.[-+0-9A-Za-z]+)*')
	WILDCARDS = ('*', 'x', 'X')
	WILDCARD = -1

	component: List[int]
	has_wildcard: bool
	pre: Optional[ExtraElement]
	build: Optional[ExtraElement]

	def __init__(self, version_str: str, allow_wildcard=True):
		"""
		:param str version_str: the version str like '1.2.3-pre4+build.5'
		"""
		if not isinstance(version_str, str):
			raise VersionParsingError('Invalid input version string')

		def separate_extra(text, char) -> Tuple[str, Optional[ExtraElement]]:
			if char in text:
				text, extra_str = text.split(char, 1)
				if not self.EXTRA_ID_PATTERN.fullmatch(extra_str):
					raise VersionParsingError('Invalid build string: ' + extra_str)
				extra = ExtraElement(extra_str)
			else:
				extra = None
			return text, extra

		self.component = []
		self.has_wildcard = False
		version_str, self.build = separate_extra(version_str, '+')
		version_str, self.pre = separate_extra(version_str, '-')
		if len(version_str) == 0:
			raise VersionParsingError('Version string is empty')
		for comp in version_str.split('.'):
			if comp in self.WILDCARDS:
				self.component.append(self.WILDCARD)
				self.has_wildcard = True
				if not allow_wildcard:
					raise VersionParsingError('Wildcard {} is not allowed'.format(comp))
			else:
				try:
					num = int(comp)
				except ValueError:
					num = None
				if num is None:
					raise VersionParsingError('Invalid version number component: {}'.format(comp))
				if num < 0:
					raise VersionParsingError('Unsupported negatived number component: {}'.format(num))
				self.component.append(num)
		if len(self.component) == 0:
			raise VersionParsingError('Empty version string')

	def __str__(self):
		version_str = '.'.join(map(lambda c: str(c) if c != self.WILDCARD else self.WILDCARDS[0], self.component))
		if self.pre is not None:
			version_str += '-' + str(self.pre)
		if self.build is not None:
			version_str += '+' + str(self.build)
		return version_str

	def __getitem__(self, index):
		if index < len(self.component):
			return self.component[index]
		else:
			return self.WILDCARD if self.component[len(self.component) - 1] == self.WILDCARD else 0

	def __lt__(self, other):
		if not isinstance(other, Version):
			raise TypeError('Cannot compare between instances of {} and {}'.format(Version.__name__, type(other).__name__))
		for i in range(max(len(self.component), len(other.component))):
			if self[i] == self.WILDCARD or other[i] == self.WILDCARD:
				continue
			if self[i] != other[i]:
				return self[i] < other[i]
		if self.pre is not None and other.pre is not None:
			return self.pre < other.pre
		elif self.pre is not None:
			return not other.has_wildcard
		elif other.pre is not None:
			return False
		else:
			return False

	def __eq__(self, other):
		return not self < other and not other < self

	def __le__(self, other):
		return self == other or self < other

	def compare_to(self, other):
		if self < other:
			return -1
		elif self > other:
			return 1
		else:
			return 0


DEFAULT_CRITERION_OPERATOR = '='


class Criterion:
	def __init__(self, opt: str, base_version: Version, criterion: Callable[[Version, Version], bool]):
		self.opt = opt
		self.base_version = base_version
		self.criterion = criterion

	def test(self, target: str or Version):
		return self.criterion(self.base_version, target)

	def __str__(self):
		return '{}{}'.format(self.opt if self.opt != DEFAULT_CRITERION_OPERATOR else '', self.base_version)


class VersionRequirement:
	CRITERIONS = {
		'<=': lambda base, ver: ver <= base,
		'>=': lambda base, ver: ver >= base,
		'<': lambda base, ver: ver < base,
		'>': lambda base, ver: ver > base,
		'=': lambda base, ver: ver == base,
		'^': lambda base, ver: ver >= base and ver[0] == base[0],
		'~': lambda base, ver: ver >= base and ver[0] == base[0] and ver[1] == base[1],
	}

	def __init__(self, requirements: str):
		if not isinstance(requirements, str):
			raise VersionParsingError('Requirements should be a str, not {}'.format(type(requirements).__name__))
		self.criterions = []  # type: List[Criterion]
		for requirement in requirements.split(' '):
			if len(requirement) > 0:
				for prefix, func in self.CRITERIONS.items():
					if requirement.startswith(prefix):
						opt = prefix
						base_version = requirement[len(prefix):]
						break
				else:
					opt = DEFAULT_CRITERION_OPERATOR
					base_version = requirement
				self.criterions.append(Criterion(opt, Version(base_version), self.CRITERIONS[opt]))

	def accept(self, version: Version or str):
		if isinstance(version, str):
			version = Version(version)
		for criterion in self.criterions:
			if not criterion.test(version):
				return False
		return True

	def __str__(self):
		return ' '.join(map(str, self.criterions))


class VersionParsingError(ValueError):
	pass



import functools
import inspect
import threading
import time
from typing import Optional, Callable

__all__ = [
	'new_thread',
	'FunctionThread'
]

from mcdreforged.utils import misc_util


class FunctionThread(threading.Thread):
	__NONE = object()

	def __init__(self, target, name, args, kwargs):
		super().__init__(target=target, args=args, kwargs=kwargs, name=name, daemon=True)
		self.__return_value = self.__NONE
		self.__error = None

		def wrapped_target(*args_, **kwargs_):
			try:
				self.__return_value = target(*args_, **kwargs_)
			except Exception as e:
				self.__error = e
				raise e from None

		self._target = wrapped_target

	def get_return_value(self, block: bool = False, timeout: Optional[float] = None):
		if block:
			self.join(timeout)
		if self.__return_value is self.__NONE:
			if self.is_alive():
				raise RuntimeError('The thread is still running')
			raise self.__error
		return self.__return_value


def new_thread(thread_name: Optional[str or Callable] = None):
	"""
	Use a new thread to execute the decorated function
	The function return value will be set to the thread instance that executes this function
	The name of the thread can be specified in parameter
	"""
	def wrapper(func):
		@functools.wraps(func)  # to preserve the origin function information
		def wrap(*args, **kwargs):
			thread = FunctionThread(target=func, args=args, kwargs=kwargs, name=thread_name)
			thread.start()
			return thread
		# bring the signature of the func to the wrap function
		# so inspect.getfullargspec(func) works correctly
		misc_util.copy_signature(wrap, func)
		wrap.original = func  # access this field to get the original function
		return wrap
	# Directly use @on_new_thread without ending brackets case
	if isinstance(thread_name, Callable):
		this_is_a_function = thread_name
		thread_name = None
		return wrapper(this_is_a_function)
	# Use @on_new_thread with ending brackets case
	return wrapper


def test():
	@new_thread()
	def bla1(value):
		time.sleep(0.5)
		print(threading.current_thread().getName() + ' ' + str(value))

	@new_thread
	def bla2(value):
		time.sleep(0.5)
		print(threading.current_thread().getName() + ' ' + str(value))

	@new_thread('awa')
	def bla3(value):
		time.sleep(0.5)
		print(threading.current_thread().getName() + ' ' + str(value))

	print(bla1, bla1('1'), inspect.getfullargspec(bla1))
	print(bla2, bla2('2'), inspect.getfullargspec(bla2))
	print(bla3, bla3('3'), inspect.getfullargspec(bla3))
	time.sleep(1)


if __name__ == '__main__':
	test()



import json
import os
import sys
from argparse import ArgumentParser
from typing import Optional
from zipfile import ZipFile, ZIP_DEFLATED

from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.utils import file_util

try:
	from mcdreforged.constants import core_constant, plugin_constant
	from mcdreforged.mcdr_server import MCDReforgedServer
except ModuleNotFoundError:
	print('It seems that you have not installed all require modules')
	raise


def environment_check():
	python_version = sys.version_info.major + sys.version_info.minor * 0.1
	if python_version < 3.6:
		print('Python 3.6+ is needed')
		raise Exception('Python version {} is too old'.format(python_version))


def entry_point():
	environment_check()
	if len(sys.argv) == 1:
		run_mcdr()
		return

	parser = ArgumentParser(
		prog=core_constant.PACKAGE_NAME,
		description='{} CLI'.format(core_constant.NAME),
	)
	parser.add_argument('-q', '--quiet', help='Disable CLI output', action='store_true')
	subparsers = parser.add_subparsers(title='Command', help='Available commands', dest='subparser_name')

	subparsers.add_parser('start', help='Start {}'.format(core_constant.NAME))
	subparsers.add_parser('init', help='Prepare the working environment of {}. Create commonly used folders and generate default configure and permission files'.format(core_constant.NAME))
	subparsers.add_parser('gendefault', help='Generate default configure and permission files at current working directory. Existed files will be overwritten')

	parser_pack = subparsers.add_parser('pack', help='Pack plugin files into a packed plugin')
	parser_pack.add_argument('-i', '--input', help='The input directory which the plugin is in, default: current directory', default='.')
	parser_pack.add_argument('-o', '--output', help='The output directory to store the zipped plugin, default: current directory', default='.')
	parser_pack.add_argument('-n', '--name', help='A specific name to the output zipped plugin file. If not given the metadata specific name or a default one will be used', default=None)

	result = parser.parse_args()
	quiet = result.quiet

	if result.subparser_name == 'start':
		run_mcdr()
	elif result.subparser_name == 'init':
		initialize_environment(quiet=quiet)
	elif result.subparser_name == 'gendefault':
		generate_default_stuffs(quiet=quiet)
	elif result.subparser_name == 'pack':
		make_packed_plugin(result.input, result.output, result.name, quiet=quiet)


def run_mcdr():
	print('{} {} is starting up'.format(core_constant.NAME, core_constant.VERSION))
	print('{} is open source, u can find it here: {}'.format(core_constant.NAME, core_constant.GITHUB_URL))
	try:
		mcdreforged_server = MCDReforgedServer()
	except:
		print('Fail to initialize {}'.format(core_constant.NAME_SHORT))
		raise
	else:
		if mcdreforged_server.is_initialized():
			mcdreforged_server.start()
		else:
			# If it's not initialized, config file or permission file is missing
			# Just dont do anything to let the user to check the files
			pass


def initialize_environment(*, quiet: bool = False):
	MCDReforgedServer(initialize_environment=True)
	if not quiet:
		print('Initialized environment for {} in {}'.format(core_constant.NAME, os.getcwd()))


def generate_default_stuffs(*, quiet: bool = False):
	MCDReforgedServer(generate_default_only=True)
	if not quiet:
		print('Generated default configure and permission files in {}'.format(os.getcwd()))


def make_packed_plugin(input_dir: str, output_dir: str, file_name: Optional[str], *, quiet: bool = False):
	writeln = print if not quiet else lambda *args, **kwargs: None

	if not os.path.isdir(input_dir):
		writeln('Invalid input directory {}'.format(input_dir))
		return
	if not os.path.isdir(output_dir):
		os.makedirs(output_dir)

	meta_file_path = os.path.join(input_dir, plugin_constant.PLUGIN_META_FILE)
	req_file_path = os.path.join(input_dir, plugin_constant.PLUGIN_REQUIREMENTS_FILE)
	if not os.path.isfile(meta_file_path):
		writeln('Plugin metadata file {} not found'.format(meta_file_path))
		return
	try:
		with open(meta_file_path, encoding='utf8') as meta_file:
			meta_dict = json.load(meta_file)  # type: dict
		assert isinstance(meta_dict, dict)
		meta = Metadata(meta_dict)
	except Exception as e:
		writeln('Fail to load plugin metadata from {}: {}'.format(meta_file_path, e))
		return
	writeln('Plugin ID: {}'.format(meta.id))
	writeln('Plugin version: {}'.format(meta.version))
	if file_name is None:
		file_name = meta.archive_name
	if file_name is None:
		file_name = '{}-v{}'.format(meta.name.replace(' ', '') or meta.id, meta.version)

	file_name = file_name.format(id=meta.id, version=meta.version)
	if file_util.get_file_suffix(file_name) not in plugin_constant.PACKED_PLUGIN_FILE_SUFFIXES:
		file_name += plugin_constant.PACKED_PLUGIN_FILE_SUFFIXES[0]
	file_counter = 0

	def write(base_path: str, *, directory_only: bool):
		nonlocal file_counter
		if os.path.isdir(base_path):
			dir_arc = os.path.basename(base_path)
			zip_file.write(base_path, arcname=dir_arc)
			file_counter += 1
			writeln('Creating directory: {} -> {}'.format(base_path, dir_arc))
			for dir_path, dir_names, file_names in os.walk(base_path):
				if os.path.basename(dir_path) == '__pycache__':
					continue
				for file_name_ in file_names + dir_names:
					full_path = os.path.join(dir_path, file_name_)
					if os.path.isdir(full_path) and os.path.basename(full_path) == '__pycache__':
						continue
					arc_name = os.path.join(dir_arc, full_path.replace(base_path, '', 1).lstrip(os.sep))
					zip_file.write(full_path, arcname=arc_name)
					file_counter += 1
					writeln('  Writing: {} -> {}'.format(full_path, arc_name))
		elif os.path.isfile(base_path) and not directory_only:
			arc_name = os.path.basename(base_path)
			zip_file.write(base_path, arcname=arc_name)
			file_counter += 1
			writeln('Writing single file: {} -> {}'.format(base_path, arc_name))
		else:
			writeln('[WARN] {} not found! ignored'.format(base_path))

	writeln('Packing plugin "{}" into "{}" ...'.format(meta.id, file_name))
	with ZipFile(os.path.join(output_dir, file_name), 'w', ZIP_DEFLATED) as zip_file:
		write(meta_file_path, directory_only=False)  # metadata
		write(req_file_path, directory_only=False)  # requirement
		write(os.path.join(input_dir, meta.id), directory_only=True)  # source module
		for resource_path in meta.resources:  # resources
			write(os.path.join(input_dir, resource_path), directory_only=False)

	writeln('Packed {} files/folders into "{}"'.format(file_counter, file_name))
	writeln('Done')



"""
Type hints are always nice to have
"""
from mcdreforged.command.command_source import CommandSource, ConsoleCommandSource, PlayerCommandSource, \
	InfoCommandSource, PluginCommandSource
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.meta.version import Version, VersionRequirement
from mcdreforged.plugin.server_interface import ServerInterface, PluginServerInterface
from mcdreforged.preference.preference_manager import PreferenceItem
from mcdreforged.utils.logger import SyncStdoutStreamHandler, MCDReforgedLogger


__all__ = [
	# Usually-used ones
	'ServerInterface', 'PluginServerInterface', 'Info',

	# Server stuffs
	'ServerInformation',

	# Command sources
	'CommandSource', 'InfoCommandSource', 'PlayerCommandSource', 'ConsoleCommandSource', 'PluginCommandSource',

	# Plugin things
	'Metadata', 'Version', 'VersionRequirement',

	# Permission
	'PermissionLevel',

	# Preference
	'PreferenceItem',

	# Logging
	'SyncStdoutStreamHandler', 'MCDReforgedLogger'
]



"""
Thread for plugin call
"""
import collections
import queue
import threading
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer
	from mcdreforged.plugin.type.plugin import AbstractPlugin


TaskData = collections.namedtuple('TaskData', 'callback plugin')


class PluginThread(threading.Thread):
	counter = 0

	def __init__(self, thread_pool, temporary=False, task=None):
		super().__init__()
		self.daemon = True
		self.thread_pool = thread_pool
		self.flag_interrupt = temporary
		self.task = task
		self.id = PluginThread.counter
		PluginThread.counter += 1
		self.original_name = '{}-idle'.format(self)
		self.setName(self.original_name)

	def run(self):
		try:
			while True:
				try:
					if self.task is not None:
						task_data = self.task
						self.task = None
					else:
						task_data = self.thread_pool.task_queue.get(timeout=0.01)  # type: TaskData
				except queue.Empty:
					pass
				else:
					plugin = task_data.plugin  # type: AbstractPlugin
					with plugin.plugin_manager.with_plugin_context(plugin):
						self.setName('{}@{}'.format(self, plugin.get_id()))
						self.thread_pool.working_count += 1
						try:
							task_data.callback()
						except:
							self.thread_pool.mcdr_server.logger.exception('Exception in thread created by {}'.format(plugin))
						finally:
							self.thread_pool.working_count -= 1
							self.setName(self.original_name)
				finally:
					if self.flag_interrupt:
						break
		finally:
			with self.thread_pool.threads_write_lock:
				self.thread_pool.threads.remove(self)

	def set_interrupt(self):
		self.flag_interrupt = True

	def join(self, timeout=None):
		self.set_interrupt()
		super().join(timeout)

	def __repr__(self):
		return '{}{}'.format(self.__class__.__name__, self.id)


class PluginThreadPool:
	def __init__(self, mcdr_server: 'MCDReforgedServer', max_thread):
		self.mcdr_server = mcdr_server
		self.threads = []
		self.threads_write_lock = threading.Lock()
		self.task_queue = queue.Queue()
		self.working_count = 0
		self.max_thread = max_thread

	def start(self):
		for i in range(self.max_thread):
			thread = PluginThread(self)
			thread.start()
			self.threads.append(thread)

	def set_max_thread(self, max_thread):
		self.max_thread = max_thread
		for i, thread in enumerate(self.threads.copy()):
			if i >= self.max_thread:
				thread.set_interrupt()

	def add_task(self, func: Callable, plugin: 'AbstractPlugin'):
		"""
		Added a task to executing the callback of a listener to the dynamic thread pool and execute it
		If the thread pool is not enough a new temporary thread will start to process the task
		"""
		task_data = TaskData(callback=func, plugin=plugin)
		if self.working_count >= self.max_thread:
			thread = PluginThread(self, temporary=True, task=task_data)
			with self.threads_write_lock:
				self.threads.append(thread)
			thread.start()
			return thread
		else:
			self.task_queue.put(task_data)

	def join(self, timeout=None):
		for thread in self.threads.copy():
			thread.join(timeout)



from abc import ABC

from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.type.plugin import AbstractPlugin


class PermanentPlugin(AbstractPlugin, ABC):
	def __init__(self, plugin_manager):
		super().__init__(plugin_manager, '**_builtin_**')
		self.__metadata = None

	def _set_metadata(self, metadata: Metadata):
		"""
		Needs to be called inside __init__ after super call
		"""
		self.__metadata = metadata

	def is_permanent(self) -> bool:
		return True

	def get_metadata(self) -> Metadata:
		return self.__metadata

	def get_fallback_metadata_id(self) -> str:
		raise RuntimeError()

	def ready(self):
		pass

	def reload(self):
		pass

	def unload(self):
		pass

	def remove(self):
		pass



import sys

from mcdreforged.constants import core_constant
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.type.permanent_plugin import PermanentPlugin


def __get_python_version():
	version = '{}.{}.{}'.format(sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
	if sys.version_info.releaselevel != 'final':
		version += '-{}.{}'.format(sys.version_info.releaselevel, sys.version_info.serial)
	return version


VERSION = __get_python_version()
METADATA = {
	'id': 'python',
	'version': VERSION,
	'name': 'Python {}'.format(VERSION),
	'author': 'Python Software Foundation',
	'link': 'https://www.python.org/'
}


class PythonPlugin(PermanentPlugin):
	def __init__(self, plugin_manager):
		super().__init__(plugin_manager)
		self._set_metadata(Metadata(METADATA, plugin=self))

	def load(self):
		self.mcdr_server.logger.info(self.mcdr_server.tr('python_plugin.info', core_constant.NAME, self.get_meta_name()))

	def __repr__(self):
		return 'PythonPlugin[version={}]'.format(METADATA['version'])



import re
from typing import Optional

from mcdreforged.command.builder.exception import UnknownArgument
from mcdreforged.command.builder.nodes.arguments import Text, QuotableText
from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource
from mcdreforged.minecraft.rtext import RText, RAction, RTextList, RColor
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand
from mcdreforged.utils import string_util


class Validator:
	@staticmethod
	def player_name(player):
		return re.fullmatch(r'\w{1,16}', player)


class PermissionCommand(SubCommand):
	def get_command_node(self) -> Literal:
		def permission_player_node():
			return QuotableText('player').suggests(lambda: self.mcdr_server.permission_manager.get_players())

		return (
			self.control_command_root({'permission', 'perm'}).
			runs(lambda src: src.reply(self.get_help_message(src, 'mcdr_command.help_message.permission'))).
			on_error(UnknownArgument, self.on_mcdr_command_unknown_argument).
			then(
				Literal('list').runs(lambda src: self.list_permission(src, None)).
				then(Text('level').runs(lambda src, ctx: self.list_permission(src, ctx['level'])))
			).
			then(Literal('set').then(permission_player_node().then(Text('level').runs(lambda src, ctx: self.set_player_permission(src, ctx['player'], ctx['level']))))).
			then(
				Literal({'query', 'q'}).runs(lambda src: self.query_self_permission(src)).
				then(permission_player_node().runs(lambda src, ctx: self.query_player_permission(src, ctx['player'])))
			).
			then(Literal({'remove', 'rm'}).then(permission_player_node().runs(lambda src, ctx: self.remove_player_permission(src, ctx['player'])))).
			then(Literal({'setdefault', 'setd'}).then(Text('level').runs(lambda src, ctx: self.set_default_permission(src, ctx['level']))))
		)

	def set_player_permission(self, source: CommandSource, player: str, value: str):
		permission_level = PermissionLevel.get_level(value)
		if permission_level is None:
			source.reply(self.tr('mcdr_command.invalid_permission_level'))
		elif not Validator.player_name(player):
			source.reply(self.tr('mcdr_command.invalid_player_name'))
		else:
			# Source with permission level x is allowed manipulate players/level in permission level range [0, x]
			if not source.has_permission(max(permission_level.level, self.mcdr_server.permission_manager.get_player_permission_level(player))):
				source.reply(self.tr('mcdr_command.permission_not_enough'))
			else:
				self.mcdr_server.permission_manager.set_permission_level(player, permission_level)
				if source.is_player:
					source.reply(self.tr('permission_manager.set_permission_level.done', player, permission_level.name))

	def query_self_permission(self, source: CommandSource):
		source.reply(self.tr('mcdr_command.query_player_permission.self', PermissionLevel.from_value(source.get_permission_level())))

	def query_player_permission(self, source: CommandSource, player: str):
		if not Validator.player_name(player):
			source.reply(self.tr('mcdr_command.invalid_player_name'))
			return
		else:
			level = self.mcdr_server.permission_manager.get_player_permission_level(player, auto_add=False)
			if level is not None:
				source.reply(self.tr('mcdr_command.query_player_permission.player', player, PermissionLevel.from_value(level)))
			else:
				source.reply(self.tr('mcdr_command.query_player_permission.player_unknown', player))

	def remove_player_permission(self, source: CommandSource, player: str):
		if not Validator.player_name(player):
			source.reply(self.tr('mcdr_command.invalid_player_name'))
		else:
			if not source.has_permission(self.mcdr_server.permission_manager.get_player_permission_level(player)):
				source.reply(self.tr('mcdr_command.permission_not_enough'))
			else:
				self.mcdr_server.permission_manager.remove_player(player)
				source.reply(self.tr('mcdr_command.remove_player_permission.player_removed', player))

	def list_permission(self, source: CommandSource, target_value: Optional[str]):
		specified_level = PermissionLevel.get_level(target_value)
		if specified_level is None:
			# show default level information if target permission not specified
			source.reply(
				RText(self.tr(
					'mcdr_command.list_permission.show_default',
					self.mcdr_server.permission_manager.get_default_permission_level()
				))
				.c(RAction.suggest_command, '{} permission setdefault '.format(self.control_command_prefix))
				.h(self.tr('mcdr_command.list_permission.suggest_setdefault'))
			)
		for permission_level in PermissionLevel.INSTANCES:
			if specified_level is None or permission_level == specified_level:
				source.reply(
					RText('§7[§e{}§7]§r'.format(permission_level.name))
					.c(RAction.run_command, '{} permission list {}'.format(self.control_command_prefix, permission_level.name))
					.h(self.tr('mcdr_command.list_permission.suggest_list', permission_level.name))
				)
				for player in self.mcdr_server.permission_manager.get_permission_group_list(permission_level.name):
					texts = RText('§7-§r {}'.format(player))
					if self.can_see_rtext(source):
						texts += RTextList(
							RText(' [✎]', color=RColor.gray)
							.c(RAction.suggest_command, '{} permission set {} '.format(self.control_command_prefix, string_util.auto_quotes(player)))
							.h(self.tr('mcdr_command.list_permission.suggest_set', player)),
							RText(' [×]', color=RColor.gray)
							.c(RAction.suggest_command, '{} permission remove {}'.format(self.control_command_prefix, string_util.auto_quotes(player)))
							.h(self.tr('mcdr_command.list_permission.suggest_disable', player)),
						)
					source.reply(texts)

	def set_default_permission(self, source: CommandSource, value: str):
		permission_level = PermissionLevel.get_level(value)
		if permission_level is None:
			source.reply(self.tr('mcdr_command.invalid_permission_level'))
		elif not source.has_permission(permission_level.level):
			source.reply(self.tr('mcdr_command.permission_not_enough'))
		else:
			self.mcdr_server.permission_manager.set_default_permission_level(permission_level)
			if source.is_player:
				source.reply(self.tr('permission_manager.set_default_permission_level.done', permission_level.name))




from mcdreforged.handler.impl.vanilla_handler import VanillaHandler


class ForgeHandler(VanillaHandler):
	# [18:26:03] [Server thread/INFO] [FML]: Unloading dimension 1
	# [18:26:03] [Server thread/INFO] [minecraft/DedicatedServer]: Done (9.855s)! For help, type "help" or "?"
	# [18:29:30] [Server thread/INFO] [minecraft/DedicatedServer]: <Steve> tests
	# [09:00:00] [Server thread/INFO]: <Steve> Hello  // vanilla format, in some old forge servers e.g. forge 1.7.10
	@classmethod
	def get_content_parsing_formatter(cls):
		return (
			'[{hour:d}:{min:d}:{sec:d}] [{thread}/{logging}] [{forge_logging}]: {content}',
			super().get_content_parsing_formatter()
		)



from mcdreforged.handler.impl.bukkit_handler import BukkitHandler
from mcdreforged.handler.impl.vanilla_handler import VanillaHandler


class Bukkit14Handler(BukkitHandler):
	# 1.14.4+ bukkit / spigot change it's console logger into vanilla like format
	# idk why they did this
	# paper is not included

	@classmethod
	def get_content_parsing_formatter(cls):
		return VanillaHandler.get_content_parsing_formatter()



import threading
from typing import Callable, Union, Optional

from mcdreforged.plugin.plugin_event import PluginEvent
from mcdreforged.plugin.server_interface import ServerInterface

__all__ = [
	'event_listener'
]


def event_listener(event: Union[PluginEvent, str], *, priority: Optional[int] = None):
	"""
	Use a new thread to execute the decorated function
	The function return value will be set to the thread instance that executes this function
	The name of the thread can be specified in parameter
	"""
	def wrapper(callback: Callable):
		server = ServerInterface.get_instance().as_plugin_server_interface()
		if server is None:
			raise RuntimeError('Cannot get current executing plugin, current thread: {}'.format(threading.current_thread()))
		server.register_event_listener(event, callback, priority)
		return callback

	if not isinstance(event, (PluginEvent, str)):
		raise TypeError('An event parameter is required')
	return wrapper



import threading

from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource
from mcdreforged.constants import core_constant
from mcdreforged.minecraft.rtext import RText, RColor, RAction, RStyle
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand


class StatusCommand(SubCommand):
	def get_command_node(self) -> Literal:
		return (
			self.public_command_root('status').runs(self.print_mcdr_status)
		)

	def print_mcdr_status(self, source: CommandSource):
		def bool_formatter(bl: bool) -> RText:
			return RText(bl, RColor.green if bl else RColor.red)
		rcon_status_dict = {
			True: self.tr('mcdr_command.print_mcdr_status.online'),
			False: self.tr('mcdr_command.print_mcdr_status.offline')
		}

		source.reply(
			RText(self.tr('mcdr_command.print_mcdr_status.line1', core_constant.NAME, core_constant.VERSION)).
			c(RAction.open_url, core_constant.GITHUB_URL).
			h(RText(core_constant.GITHUB_URL, styles=RStyle.underlined, color=RColor.blue))
		)

		if not source.has_permission(PermissionLevel.MCDR_CONTROL_LEVEL):
			return
		source.reply(RText.join('\n', [
			RText(self.tr('mcdr_command.print_mcdr_status.line2', self.tr(self.mcdr_server.mcdr_state.value))),
			RText(self.tr('mcdr_command.print_mcdr_status.line3', self.tr(self.mcdr_server.server_state.value))),
			RText(self.tr('mcdr_command.print_mcdr_status.line4', bool_formatter(self.mcdr_server.is_server_startup()))),
			RText(self.tr('mcdr_command.print_mcdr_status.line5', bool_formatter(self.mcdr_server.should_exit_after_stop()))),
			RText(self.tr('mcdr_command.print_mcdr_status.line6', rcon_status_dict[self.server_interface.is_rcon_running()])),
			RText(self.tr('mcdr_command.print_mcdr_status.line7', self.mcdr_server.plugin_manager.get_plugin_amount())).c(RAction.suggest_command, '!!MCDR plugin list')
		]))

		if not source.has_permission(PermissionLevel.PHYSICAL_SERVER_CONTROL_LEVEL):
			return
		source.reply(RText.join('\n', [
			self.tr('mcdr_command.print_mcdr_status.extra_line1', self.mcdr_server.process.pid if self.mcdr_server.process is not None else '§rN/A§r'),
			self.tr('mcdr_command.print_mcdr_status.extra_line2', self.mcdr_server.task_executor.task_queue.qsize(), core_constant.MAX_TASK_QUEUE_SIZE),
			self.tr('mcdr_command.print_mcdr_status.extra_line3', threading.active_count())
		]))
		thread_pool_counts = 0
		for thread in threading.enumerate():
			name = thread.getName()
			if not name.startswith('ThreadPoolExecutor-'):
				source.reply('  §7-§r {}'.format(name))
			else:
				thread_pool_counts += 1
		if thread_pool_counts > 0:
			source.reply('  §7-§r ThreadPoolExecutor thread x{}'.format(thread_pool_counts))



"""
Information of a plugin
"""
import re
from typing import List, Dict, TYPE_CHECKING, Optional, Union

from mcdreforged.minecraft.rtext import RTextBase, RText
from mcdreforged.plugin.meta.version import Version, VersionParsingError, VersionRequirement
from mcdreforged.translation.translation_text import RTextMCDRTranslation
from mcdreforged.utils import translation_util, misc_util
from mcdreforged.utils.types import TranslationLanguageDict

if TYPE_CHECKING:
	from mcdreforged.plugin.type.plugin import AbstractPlugin


class Metadata:
	id: str
	version: Version
	name: str
	description: Optional[Union[str, TranslationLanguageDict]]  # translation: lang -> description
	author: Optional[List[str]]
	link: Optional[str]
	dependencies: Dict[str, VersionRequirement]

	entrypoint: str
	archive_name: Optional[str]
	resources: Optional[List[str]]

	FALLBACK_VERSION = '0.0.0'

	def __init__(self, data: Optional[dict], *, plugin: Optional['AbstractPlugin'] = None):
		"""
		:param AbstractPlugin plugin: the plugin which this metadata is belong to
		:param dict or None data: a dict with information of the plugin
		"""
		if not isinstance(data, dict):
			data = {}

		def warn(*args, **kwargs):
			if plugin is not None:
				plugin.mcdr_server.logger.warning(*args, **kwargs)

		plugin_name_text = repr(plugin)

		use_fallback_id_reason = None
		self.id = data.get('id')
		if self.id is None:
			use_fallback_id_reason = 'Plugin ID of {} not found'.format(plugin_name_text)
		elif not isinstance(self.id, str) or re.fullmatch(r'[a-z0-9_]{1,64}', self.id) is None:
			use_fallback_id_reason = 'Plugin ID "{}" of {} is invalid'.format(self.id, plugin_name_text)
		if use_fallback_id_reason is not None:
			if plugin is not None:
				self.id = plugin.get_fallback_metadata_id()
				warn('{}, use fallback id {} instead'.format(use_fallback_id_reason, self.id))
			else:
				raise ValueError('Plugin id not found in metadata')
		misc_util.check_type(self.id, str)

		self.name = data.get('name', self.id)
		if isinstance(self.name, RTextBase):
			self.name = self.name.to_plain_text()
		misc_util.check_type(self.name, str)

		description = data.get('description')
		if isinstance(description, RTextBase):
			description = description.to_plain_text()
		self.description = description
		misc_util.check_type(self.description, [type(None), str, dict])

		self.author = data.get('author')
		if isinstance(self.author, str):
			self.author = [self.author]
		if isinstance(self.author, list):
			for i in range(len(self.author)):
				self.author[i] = str(self.author[i])
			if len(self.author) == 0:
				self.author = None
		misc_util.check_type(self.author, [type(None), list])

		self.link = data.get('link')
		misc_util.check_type(self.link, [type(None), str])

		version_str = data.get('version')
		if version_str:
			try:
				self.version = Version(version_str, allow_wildcard=False)
			except VersionParsingError as e:
				warn('Version "{}" of {} is invalid ({}), ignore and use fallback version instead {}'.format(version_str, plugin_name_text, e, self.FALLBACK_VERSION))
				version_str = None
		else:
			warn('{} doesn\'t specific a version, use fallback version {}'.format(plugin_name_text, self.FALLBACK_VERSION))
		if version_str is None:
			self.version = Version(self.FALLBACK_VERSION)

		self.dependencies = {}
		for plugin_id, requirement in data.get('dependencies', {}).items():
			try:
				self.dependencies[plugin_id] = VersionRequirement(requirement)
			except VersionParsingError as e:
				warn('Dependency "{}: {}" of {} is invalid ({}), ignore'.format(
					plugin_id, requirement, plugin_name_text, e
				))

		self.entrypoint = data.get('entrypoint', self.id)
		misc_util.check_type(self.entrypoint, str)
		# entrypoint module should be inside the plugin module
		if self.entrypoint != self.id and not self.entrypoint.startswith(self.id + '.'):
			raise ValueError('Invalid entry point "{}" for plugin id "{}"'.format(self.entrypoint, self.id))

		self.archive_name = data.get('archive_name')
		self.resources = data.get('resources', [])
		misc_util.check_type(self.archive_name, [type(None), str])
		misc_util.check_type(self.resources, list)

	def __repr__(self):
		return misc_util.represent(self)

	def get_description(self, lang: Optional[str] = None) -> Optional[str]:
		"""
		Get translated description str
		"""
		if isinstance(self.description, str):
			return self.description
		return translation_util.translate_from_dict(self.description, lang, default=None)

	def get_description_rtext(self) -> RTextBase:
		if isinstance(self.description, str):
			return RText(self.description)
		return RTextMCDRTranslation.from_translation_dict(self.description)


__SAMPLE_METADATA = {
	'id': 'example_plugin',   # If missing it will be the file name without .py suffix
	'version': '1.0.0',       # If missing it will be '0.0.0'
	'name': 'Sample Plugin',
	# single string description is also supported
	# 'description': 'Sample plugin for MCDR',
	'description': {
		'en_us': 'Sample plugin for MCDR'
	},
	'author': [
		'Fallen_Breath'
	],
	'link': 'https://github.com/Fallen-Breath/MCDReforged',
	'dependencies': {
		'mcdreforged': '>=1.0.0'
	},

	# Fields for packed plugins
	'entrypoint': 'example_plugin.entry',
	'archive_name': 'MyExamplePlugin-v{version}',
	'resources': [
		'my_resource_folder',
		'another_resource_file',
	]
}



from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand


class CheckUpdateCommand(SubCommand):
	def get_command_node(self) -> Literal:
		return (
			self.control_command_root({'checkupdate', 'cu'}).
			runs(lambda src: self.mcdr_server.update_helper.check_update(condition_check=lambda: True, reply_func=src.reply))
		)



import collections
from enum import unique, Enum, auto
from typing import Dict, List, TYPE_CHECKING, Tuple

from mcdreforged.plugin.meta.version import VersionRequirement
from mcdreforged.utils.logger import DebugOption

if TYPE_CHECKING:
	from mcdreforged.plugin.plugin_manager import PluginManager


class DependencyError(Exception):
	pass


class DependencyParentFail(DependencyError):
	pass


class DependencyNotFound(DependencyError):
	pass


class DependencyNotMet(DependencyError):
	pass


class DependencyLoop(DependencyError):
	pass


@unique
class VisitingState(Enum):
	UNVISITED = auto()
	PASS = auto()
	FAIL = auto()


WalkResult = collections.namedtuple('WalkResult', 'plugin_id success reason')


class DependencyWalker:
	def __init__(self, plugin_manager: 'PluginManager'):
		self.plugin_manager = plugin_manager
		self.tr = plugin_manager.mcdr_server.tr
		self.visiting_state = {}  # type: Dict[str, VisitingState]
		self.visiting_plugins = set()
		self.visiting_plugin_stack = []
		self.topo_order = []

	def get_visiting_status(self, plugin_id):
		value = self.visiting_state.get(plugin_id)
		if value is None:
			value = VisitingState.UNVISITED
			self.visiting_state[plugin_id] = value
		return value

	def ensure_loaded(self, plugin_id: str, requirement: VersionRequirement or None):
		visiting_status = self.get_visiting_status(plugin_id)
		if visiting_status == VisitingState.FAIL:
			raise DependencyParentFail(self.tr('dependency_walker.dependency_parent_failed', plugin_id))

		if plugin_id in self.visiting_plugins:
			start = 0
			for i, pid in enumerate(self.visiting_plugin_stack):
				if pid == plugin_id:
					start = i
					break
			loop_list = self.visiting_plugin_stack[start:]
			loop_list.append(plugin_id)
			loop_message = ' -> '.join(loop_list)
			raise DependencyLoop(self.tr('dependency_walker.dependency_loop', plugin_id, loop_message))

		self.visiting_plugins.add(plugin_id)
		self.visiting_plugin_stack.append(plugin_id)
		try:
			plugin = self.plugin_manager.get_plugin_from_id(plugin_id)
			if plugin is None:
				raise DependencyNotFound(self.tr('dependency_walker.dependency_not_found', plugin_id))
			plugin_name_display = plugin.get_name()
			plugin_version = plugin.get_metadata().version
			plugin_dependencies = plugin.get_metadata().dependencies

			if requirement is not None and isinstance(requirement, VersionRequirement) and not requirement.accept(plugin_version):
				raise DependencyNotMet(self.tr('dependency_walker.dependency_not_met', plugin_name_display, requirement))

			if visiting_status == VisitingState.PASS:  # no need to do further dependencies check, already done
				return

			for dep_id, req in plugin_dependencies.items():
				try:
					self.ensure_loaded(dep_id, req)
				except DependencyError as e:
					self.visiting_state[plugin_id] = VisitingState.FAIL
					self.plugin_manager.logger.debug('Set visiting state of {} to FAIL due to "{}"'.format(plugin_id, e), option=DebugOption.PLUGIN)
					raise
			if not plugin.is_permanent():
				self.topo_order.append(plugin_id)
			self.visiting_state[plugin_id] = VisitingState.PASS
		finally:
			self.visiting_plugins.remove(plugin_id)
			self.visiting_plugin_stack.pop(len(self.visiting_plugin_stack) - 1)

	def walk(self) -> List[WalkResult]:
		self.visiting_state.clear()
		self.visiting_plugins.clear()
		self.visiting_plugin_stack.clear()
		self.topo_order.clear()
		fail_list = []  # type: List[Tuple[str, DependencyError]]
		for plugin in self.plugin_manager.get_regular_plugins():
			try:
				plugin_id = plugin.get_id()
				visiting_state = self.get_visiting_status(plugin_id)
				if visiting_state is not VisitingState.FAIL:
					self.ensure_loaded(plugin_id, None)
				else:
					raise DependencyError(self.tr('dependency_walker.dependency_already_failed', plugin, visiting_state))
			except DependencyError as e:
				fail_list.append((plugin.get_id(), e))
		result = []
		for plugin_id, error in fail_list:
			result.append(WalkResult(plugin_id, False, error))
		for plugin_id in self.topo_order:
			result.append(WalkResult(plugin_id, True, None))
		return result



from mcdreforged.info_reactor.impl.general_reactor import GeneralReactor
from mcdreforged.info_reactor.impl.player_reactor import PlayerReactor
from mcdreforged.info_reactor.impl.server_reactor import ServerReactor

__all__ = [
	'GeneralReactor',
	'PlayerReactor',
	'ServerReactor',
]



import functools
import json
import os
import time
from typing import Callable, TYPE_CHECKING, Tuple, Any, Union, Optional, List, IO, Dict, Type, TypeVar

from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource, PluginCommandSource
from mcdreforged.constants import plugin_constant
from mcdreforged.info_reactor.info import Info
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.mcdr_state import MCDReforgedFlag
from mcdreforged.permission.permission_level import PermissionLevel, PermissionParam
from mcdreforged.plugin import plugin_factory
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.operation_result import SingleOperationResult, PluginOperationResult
from mcdreforged.plugin.plugin_event import EventListener, LiteralEvent, PluginEvent, MCDRPluginEvents
from mcdreforged.plugin.plugin_registry import DEFAULT_LISTENER_PRIORITY, HelpMessage
from mcdreforged.plugin.type.multi_file_plugin import MultiFilePlugin
from mcdreforged.plugin.type.plugin import AbstractPlugin
from mcdreforged.preference.preference_manager import PreferenceItem
from mcdreforged.translation.translation_text import RTextMCDRTranslation
from mcdreforged.utils import misc_util, file_util
from mcdreforged.utils.exception import IllegalCallError
from mcdreforged.utils.logger import MCDReforgedLogger, DebugOption
from mcdreforged.utils.serializer import Serializable
from mcdreforged.utils.types import MessageText, TranslationKeyDictRich, TranslationKeyDictNested

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer
	from mcdreforged.handler.abstract_server_handler import AbstractServerHandler
	from mcdreforged.plugin.plugin_manager import PluginManager
	from mcdreforged.plugin.type.regular_plugin import RegularPlugin


SerializableType = TypeVar('SerializableType')


class ServerInterface:
	"""
	An interface class for plugins to control the server
	"""

	MCDR = True  # Identifier for plugins
	__global_instance: Optional['ServerInterface'] = None  # For singleton instance storage

	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self._mcdr_server = mcdr_server
		if type(self) is ServerInterface:
			# singleton, should only occurs during MCDReforgedServer construction
			if ServerInterface.__global_instance is not None:
				self._mcdr_server.logger.warning('Double assigning the singleton instance in {}'.format(self.__class__.__name__), stack_info=True)
			ServerInterface.__global_instance = self

	@functools.lru_cache(maxsize=512, typed=True)
	def __get_logger(self, plugin_id: str):
		return MCDReforgedLogger(plugin_id)

	@property
	def logger(self) -> MCDReforgedLogger:
		"""
		Yes, a nice logger for you to use
		"""
		plugin = self._mcdr_server.plugin_manager.get_current_running_plugin()
		if plugin is not None:
			return self.__get_logger(plugin.get_id())
		else:
			return self._mcdr_server.logger

	# ------------------------
	#          Utils
	# ------------------------

	@classmethod
	def get_instance(cls) -> Optional['ServerInterface']:
		"""
		A class method, for plugins to get a ServerInterface instance anywhere as long as MCDR is running
		"""
		return cls.__global_instance

	def tr(self, translation_key: str, *args, language: Optional[str] = None, **kwargs) -> MessageText:
		"""
		Return a translated text corresponded to the translation key and format the text with given args and kwargs
		If args or kwargs contains RText element, then the result will be a RText, otherwise the result will be a regular str
		If the translation key is not recognized, the return value will be the translation key itself
		:param translation_key: The key of the translation
		:param args: The args to be formatted
		:param language: Specific language to be used in this translation, or the language that MCDR is using will be used
		:param kwargs: The kwargs to be formatted
		"""
		return self._mcdr_server.tr(translation_key, *args, language=language, **kwargs)

	def rtr(self, translation_key: str, *args, **kwargs) -> RTextMCDRTranslation:
		"""
		Return a RText derived component RTextMCDRTranslation, that only translates itself right before displaying or serializing
		Using this method instead of tr() allows you to display your texts in user's preferred language automatically
		Of course you can construct RTextMCDRTranslation yourself instead of using this method if you want
		:param translation_key: The key of the translation
		:param args: The args to be formatted
		:param kwargs: The kwargs to be formatted
		"""
		text = RTextMCDRTranslation(translation_key, *args, **kwargs)
		text.set_translator(self.tr)  # not that necessary tbh, just in case self.tr != ServerInterface.get_instance().tr somehow
		return text

	def as_basic_server_interface(self) -> 'ServerInterface':
		"""
		Return a ServerInterface instance. The type of the return value is exactly the ServerInterface
		It's used for removing the plugin information inside PluginServerInterface when you need to send a ServerInterface
		"""
		return self.get_instance()

	def as_plugin_server_interface(self) -> Optional['PluginServerInterface']:
		"""
		Return a PluginServerInterface instance. If current thread is not a MCDR provided thread and the object is not
		a PluginServerInterface instance, it will return None
		"""
		plugin = self._mcdr_server.plugin_manager.get_current_running_plugin()
		if plugin is not None:
			return plugin.server_interface
		return None

	# ------------------------
	#      Server Control
	# ------------------------

	def start(self) -> bool:
		"""
		Start the server
		:return: If the action succeed it's True. If the server is running or being starting by other plugin return False
		"""
		return self._mcdr_server.start_server()

	def stop(self) -> None:
		"""
		Soft shutting down the server by sending the correct stop command to the server
		MCDR will keep running unless exit() is invoked
		"""
		self._mcdr_server.remove_flag(MCDReforgedFlag.EXIT_AFTER_STOP)
		self._mcdr_server.stop(forced=False)

	def kill(self) -> None:
		"""
		Kill the server process group
		MCDR will keep running unless exit() is invoked
		"""
		self._mcdr_server.stop(forced=True)

	def wait_for_start(self) -> None:
		"""
		Wait until the server is able to start. In other words, wait until the server is stopped
		"""
		while self.is_server_running():
			time.sleep(0.01)

	def restart(self) -> None:
		"""
		Restart the server
		It will first soft stop the server and then wait until the server is stopped, then start the server up
		"""
		if self.is_server_running():
			self.stop()
			self.wait_for_start()
			self.start()

	def stop_exit(self) -> None:
		"""
		Soft stop the server and exit MCDR
		"""
		self._mcdr_server.stop(forced=False)

	def exit(self) -> None:
		"""
		Exit MCDR when the server is stopped
		Basically it's the same to invoking set_exit_after_stop_flag(True), but with an extra server not running check
		:raise: IllegalCallError, if the server is not stopped
		"""
		if self._mcdr_server.is_server_running():
			raise IllegalCallError('Cannot exit MCDR when the server is running')
		self._mcdr_server.with_flag(MCDReforgedFlag.EXIT_AFTER_STOP)

	def set_exit_after_stop_flag(self, flag_value: bool) -> None:
		"""
		Set the flag that indicating if MCDR should exit when the server has stopped
		If set to true, after the server stops MCDR will exit, otherwise (set to false) MCDR will just keep running
		The flag value will be set to true everything when the server starts
		The flag value is displayed in line 5 in command `!!MCDR status`
		"""
		if flag_value:
			self._mcdr_server.with_flag(MCDReforgedFlag.EXIT_AFTER_STOP)
		else:
			self._mcdr_server.remove_flag(MCDReforgedFlag.EXIT_AFTER_STOP)

	def is_server_running(self) -> bool:
		"""
		Return if the server is running
		"""
		return self._mcdr_server.is_server_running()

	def is_server_startup(self) -> bool:
		"""
		Return if the server has started up
		"""
		return self._mcdr_server.is_server_startup()

	def is_rcon_running(self) -> bool:
		"""
		Return if MCDR's rcon is running
		"""
		return self._mcdr_server.rcon_manager.is_running()

	def get_server_pid(self) -> Optional[int]:
		"""
		Return the pid of the server process
		Notes the process with this pid is a bash process, which is the parent process of real server process
		you might be interested in
		:return: The pid of the server. None if the server is stopped
		"""
		if self._mcdr_server.process is not None:
			return self._mcdr_server.process.pid
		return None

	def get_server_information(self) -> ServerInformation:
		"""
		Return a ServerInformation object indicating the information of the current server, interred from the output of the server
		Field(s) might be None if the server is offline, or the related information has not been parsed
		:return: a ServerInformation object
		"""
		return self._mcdr_server.server_information.copy()

	# ------------------------
	#     Text Interaction
	# ------------------------

	def execute(self, text: str, *, encoding: Optional[str] = None) -> None:
		"""
		Execute a command by sending the command content to server's standard input stream
		:param str text: The content of the command you want to send
		:param str encoding: The encoding method for the text
		"""
		self.logger.debug('Sending command "{}"'.format(text), option=DebugOption.PLUGIN)
		self._mcdr_server.send(text, encoding=encoding)

	@property
	def __server_handler(self) -> 'AbstractServerHandler':
		return self._mcdr_server.server_handler_manager.get_current_handler()

	def tell(self, player: str, text: MessageText, *, encoding: Optional[str] = None) -> None:
		"""
		Use command like /tellraw to send the message to the specific player
		:param player: The name of the player you want to tell
		:param text: the message you want to send to the player
		:param encoding: The encoding method for the text
		"""
		with RTextMCDRTranslation.language_context(self._mcdr_server.preference_manager.get_preferred_language(player)):
			command = self.__server_handler.get_send_message_command(player, text, self.get_server_information())
		if command is not None:
			self.execute(command, encoding=encoding)

	def say(self, text: MessageText, *, encoding: Optional[str] = None) -> None:
		"""
		Use command like /tellraw @a to broadcast the message in game
		:param text: the message you want to send
		:param encoding: The encoding method for the text
		"""
		command = self.__server_handler.get_broadcast_message_command(text, self.get_server_information())
		if command is not None:
			self.execute(command, encoding=encoding)

	def broadcast(self, text: MessageText, *, encoding: Optional[str] = None) -> None:
		"""
		Broadcast the message in game and to the console
		:param text: the message you want to send
		:param encoding: The encoding method for the text
		"""
		self.say(text, encoding=encoding)
		misc_util.print_text_to_console(self.logger, text)

	def reply(self, info: Info, text: MessageText, *, encoding: Optional[str] = None, console_text: Optional[MessageText] = None):
		"""
		Reply to the source of the Info
		If the Info is from a player then use tell to reply the player
		Otherwise if the Info is from the console use logger.info to output to the console
		In the rest of the situations, the Info is not from a user, a IllegalCallError is raised
		:param info: the Info you want to reply to
		:param text: the message you want to send
		:param console_text: If it's specified, console_text will be used instead of text when replying to console
		:param encoding: The encoding method for the text
		"""
		source = info.get_command_source()
		if source is None:
			raise IllegalCallError('Cannot reply to the given info instance')
		if not source.is_console:
			if console_text is not None:
				text = console_text
		source.reply(text, encoding=encoding)

	# ------------------------
	#      Plugin Queries
	# ------------------------

	def __existed_plugin_info_getter(self, plugin_id: str, handler: Callable[['AbstractPlugin'], Any], *, regular: bool):
		if regular:
			plugin = self._mcdr_server.plugin_manager.get_regular_plugin_from_id(plugin_id)
		else:
			plugin = self._mcdr_server.plugin_manager.get_plugin_from_id(plugin_id)
		if plugin is not None:
			return handler(plugin)
		return None

	def get_plugin_metadata(self, plugin_id: str) -> Optional[Metadata]:
		"""
		Return the metadata of the specified plugin, or None if the plugin doesn't exist
		:param plugin_id: The plugin id of the plugin to query metadata
		"""
		return self.__existed_plugin_info_getter(plugin_id, lambda plugin: plugin.get_metadata(), regular=False)

	def get_plugin_file_path(self, plugin_id: str) -> Optional[str]:
		"""
		Return the file path of the specified plugin, or None if the plugin doesn't exist
		:param plugin_id: The plugin id of the plugin to query file path
		"""
		return self.__existed_plugin_info_getter(plugin_id, lambda plugin: plugin.plugin_path, regular=False)

	def get_plugin_instance(self, plugin_id: str) -> Optional[Any]:
		"""
		Return the entrypoint module instance of the specific plugin, or None if the plugin doesn't exist
		:param plugin_id: The plugin id of the plugin you want
		:return: A entrypoint module instance, or None if the plugin doesn't exist
		"""
		return self.__existed_plugin_info_getter(plugin_id, lambda plugin: plugin.entry_module_instance, regular=True)

	def get_plugin_list(self) -> List[str]:
		"""
		Return a list containing all loaded plugin id like ["my_plugin", "another_plugin"]
		"""
		return [plugin.get_id() for plugin in self._mcdr_server.plugin_manager.get_regular_plugins()]

	def __get_files_in_plugin_directories(self) -> List[str]:
		result = []
		for plugin_directory in self._mcdr_server.plugin_manager.plugin_directories:
			result.extend(file_util.list_all(plugin_directory))
		return result

	def get_unloaded_plugin_list(self) -> List[str]:
		"""
		Return a list containing all unloaded plugin file path like ["plugins/MyPlugin.mcdr"]
		"""
		return list(filter(
			lambda file_path: not self._mcdr_server.plugin_manager.contains_plugin_file(file_path) and plugin_factory.is_plugin(file_path),
			self.__get_files_in_plugin_directories()
		))

	def get_disabled_plugin_list(self) -> List[str]:
		"""
		Return a list containing all disabled plugin file path like ["plugins/MyPlugin.mcdr.disabled"]
		"""
		return list(filter(
			lambda file_path: plugin_factory.is_disabled_plugin(file_path),
			self.__get_files_in_plugin_directories()
		))

	def get_all_metadata(self) -> Dict[str, Metadata]:
		"""
		Return a dict containing metadatas of all loaded plugin with (plugin_id, metadata) as key-value pair
		"""
		result = {}
		for plugin in self._mcdr_server.plugin_manager.get_all_plugins():
			result[plugin.get_id()] = plugin.get_metadata()
		return result

	# ------------------------
	#     Plugin Operations
	# ------------------------

	# Notes: All plugin manipulation will trigger a dependency check, which might cause unwanted plugin operations

	def __check_if_success(self, operation_result: SingleOperationResult, check_loaded: bool) -> bool:
		"""
		Check if there's any plugin inside the given operation result (load result / reload result etc.)
		Then check if the plugin passed the dependency check if param check_loaded is True
		"""
		success = operation_result.has_success()
		if success and check_loaded:
			plugin = operation_result.success_list[0]
			success = plugin in self._mcdr_server.plugin_manager.last_operation_result.dependency_check_result.success_list
		return success

	def __not_loaded_regular_plugin_manipulate(self, plugin_file_path: str, handler: Callable[['PluginManager'], Callable[[str], Any]]) -> bool:
		"""
		Manipulate a not loaded regular plugin from a given file path
		:param plugin_file_path: The path to the not loaded new plugin
		:param handler: What you want to do with Plugin Manager to the given file path
		:return: If success
		"""
		self._mcdr_server.plugin_manager.last_operation_result.clear()
		handler(self._mcdr_server.plugin_manager)(plugin_file_path)
		return self.__check_if_success(self._mcdr_server.plugin_manager.last_operation_result.load_result, check_loaded=True)  # the operations is always loading a plugin

	def __existed_regular_plugin_manipulate(self, plugin_id: str, handler: Callable[['PluginManager'], Callable[['RegularPlugin'], Any]], result_getter: Callable[[PluginOperationResult], SingleOperationResult], check_loaded: bool) -> bool or None:
		"""
		Manipulate a loaded regular plugin from a given plugin id
		:param plugin_id: The plugin id of the plugin you want to manipulate
		:param handler: What callable you want to use with Plugin Manager to the plugin id,
		the returned callable accepts the plugin instance
		:param result_getter: How to get the single operation result from the plugin operation result.
		It's used to determine if the operation succeeded
		:return: If success, None if plugin not found
		"""
		plugin = self._mcdr_server.plugin_manager.get_regular_plugin_from_id(plugin_id)
		if plugin is not None:
			self._mcdr_server.plugin_manager.last_operation_result.clear()
			handler(self._mcdr_server.plugin_manager)(plugin)
			opt_result = result_getter(self._mcdr_server.plugin_manager.last_operation_result)
			return self.__check_if_success(opt_result, check_loaded)
		return None

	def load_plugin(self, plugin_file_path: str) -> bool:
		"""
		Load a plugin from the given file path
		:param plugin_file_path: The file path of the plugin to load. Example: "plugins/my_plugin.py"
		:return: If the plugin gets loaded successfully
		"""
		return self.__not_loaded_regular_plugin_manipulate(plugin_file_path, lambda mgr: mgr.load_plugin)

	def enable_plugin(self, plugin_file_path: str) -> bool:
		"""
		Enable an unloaded plugin from the given path
		:param plugin_file_path: The file path of the plugin to enable. Example: "plugins/my_plugin.py.disabled"
		:return: If the plugin gets enabled successfully
		"""
		return self.__not_loaded_regular_plugin_manipulate(plugin_file_path, lambda mgr: mgr.enable_plugin)

	def reload_plugin(self, plugin_id: str) -> Optional[bool]:
		"""
		Reload a plugin specified by plugin id
		:param plugin_id: The id of the plugin to reload. Example: "my_plugin"
		:return: A bool indicating if the plugin gets reloaded successfully, or None if plugin not found
		"""
		return self.__existed_regular_plugin_manipulate(plugin_id, lambda mgr: mgr.reload_plugin, lambda lor: lor.reload_result, check_loaded=True)

	def unload_plugin(self, plugin_id: str) -> Optional[bool]:
		"""
		Unload a plugin specified by plugin id
		:param plugin_id: The id of the plugin to unload. Example: "my_plugin"
		:return: A bool indicating if the plugin gets unloaded successfully, or None if plugin not found
		"""
		return self.__existed_regular_plugin_manipulate(plugin_id, lambda mgr: mgr.unload_plugin, lambda lor: lor.unload_result, check_loaded=False)

	def disable_plugin(self, plugin_id: str) -> Optional[bool]:
		"""
		Disable an unloaded plugin from the given path
		:param plugin_id: The id of the plugin to disable. Example: "my_plugin"
		:return: A bool indicating if the plugin gets disabled successfully, or None if plugin not found
		"""
		return self.__existed_regular_plugin_manipulate(plugin_id, lambda mgr: mgr.disable_plugin, lambda lor: lor.unload_result, check_loaded=False)

	def refresh_all_plugins(self) -> None:
		"""
		Reload all plugins, load all new plugins and then unload all removed plugins
		"""
		self._mcdr_server.plugin_manager.refresh_all_plugins()

	def refresh_changed_plugins(self) -> None:
		"""
		Reload all changed plugins, load all new plugins and then unload all removed plugins
		"""
		self._mcdr_server.plugin_manager.refresh_changed_plugins()

	def dispatch_event(self, event: PluginEvent, args: Tuple[Any, ...], *, on_executor_thread: bool = True) -> None:
		"""
		Dispatch an event to all loaded plugins
		The event will be immediately dispatch if it's on the task executor thread, or gets enqueued if it's on other thread
		:param event: The event to dispatch. It need to be a PluginEvent instance. For simple usage, you can create a
		LiteralEvent instance for this argument
		:param args: The argument that will be used to invoke the event listeners. An ServerInterface instance will be
		automatically added to the beginning of the argument list
		:param on_executor_thread: By default the event will be dispatched in a new task in task executor thread
		If it's set to false. The event will be dispatched immediately
		"""
		misc_util.check_type(event, PluginEvent)
		if MCDRPluginEvents.contains_id(event.id):
			raise ValueError('Cannot dispatch event with already exists event id {}'.format(event.id))
		self._mcdr_server.plugin_manager.dispatch_event(event, args, on_executor_thread=on_executor_thread)

	# ------------------------
	#        Permission
	# ------------------------

	def get_permission_level(self, obj: Union[str, Info, CommandSource]) -> int:
		"""
		Return an int indicating permission level number the given object has
		The object could be a str indicating the name of a player, an Info instance or a command source
		:param obj: The object your are querying
		:raise: TypeError, if the type of the given object is not supported for permission querying
		"""
		if isinstance(obj, Info):  # Info instance
			obj = obj.get_command_source()
			if obj is None:
				raise TypeError('The Info instance is not from a user')
		if isinstance(obj, CommandSource):  # Command Source
			return obj.get_permission_level()
		elif isinstance(obj, str):  # Player name
			return self._mcdr_server.permission_manager.get_player_permission_level(obj)
		else:
			raise TypeError('Unsupported permission level querying for type {}'.format(type(obj)))

	def set_permission_level(self, player: str, value: PermissionParam) -> None:
		"""
		Set the permission level of the given player
		:param player: The name of the player that you want to set his/her permission level
		:param value: The target permission level you want to set the player to. It can be an int or a str as long as
		it's related to the permission level. Available examples: 1, '1', 'user'
		:raise: TypeError if the value parameter doesn't proper represent a permission level
		"""
		level = PermissionLevel.get_level(value)
		if level is None:
			raise TypeError('Parameter level needs to be a permission related value')
		self._mcdr_server.permission_manager.set_permission_level(player, level)

	# ------------------------
	#         Command
	# ------------------------

	def get_plugin_command_source(self) -> PluginCommandSource:
		"""
		Return a simple plugin command source for e.g. command execution
		It's not player or console, it has maximum permission level, it use ServerInterface.logger for replying
		"""
		return PluginCommandSource(self, None)

	def execute_command(self, command: str, source: CommandSource = None) -> None:
		"""
		Execute a single command using the command system of MCDR
		:param str command: The command you want to execute
		:param CommandSource source: The command source that is used to execute the command. If it's not specified MCDR
		will use ServerInterface.get_plugin_command_source as fallback command source
		"""
		if source is None:
			source = self.get_plugin_command_source()
		misc_util.check_type(command, str)
		misc_util.check_type(source, CommandSource)
		self._mcdr_server.command_manager.execute_command(command, source)

	# ------------------------
	#        Preference
	# ------------------------

	def get_preference(self, obj: Union[str, CommandSource]) -> PreferenceItem:
		"""
		Return the MCDR preference of the given object. The object can be a str indicating the name of a player, or a
		command source. For command source, only PlayerCommandSource and ConsoleCommandSource are supported
		:param obj: The object to querying preference
		:raise: TypeError, if the type of the given object is not supported for preference querying
		"""
		pref = self._mcdr_server.preference_manager.get_preference(obj, strict_type_check=True)
		return PreferenceItem.deserialize(pref.serialize())  # make a copy

	# ------------------------
	#           Misc
	# ------------------------

	def is_on_executor_thread(self) -> bool:
		"""
		Return if the current thread is the task executor thread
		Task executor thread is the main thread to parse messages and trigger listeners where some ServerInterface APIs
		are required to be invoked on
		"""
		return self._mcdr_server.task_executor.is_on_thread()

	def rcon_query(self, command: str) -> Optional[str]:
		"""
		Send command to the server through rcon connection
		:param str command: The command you want to send to the rcon server
		:return: The result that server returned from rcon. Return None if rcon is not running or rcon query failed
		"""
		return self._mcdr_server.rcon_manager.send_command(command)

	def get_mcdr_language(self) -> str:
		"""
		Return the current language MCDR is using
		"""
		return self._mcdr_server.translation_manager.language

	def get_mcdr_config(self) -> dict:
		"""
		Return the current config of MCDR as a dict
		"""
		return self._mcdr_server.config.to_dict()

	def schedule_task(self, callable_: Callable[[], Any], *, block: bool = False, timeout: Optional[float] = None) -> None:
		"""
		Schedule a callback task to be run in task executor thread
		:param callable_: The callable object to be run. It should accept 0 parameter
		:param block: If blocks until the callable finished execution
		:param timeout: The timeout of the blocking operation if block=True
		"""
		self._mcdr_server.task_executor.add_regular_task(callable_, block=block, timeout=timeout)


class PluginServerInterface(ServerInterface):
	"""
	An extended ServerInterface class that adds the ability for plugins to control the plugin itself
	"""

	def __init__(self, mcdr_server: 'MCDReforgedServer', plugin: AbstractPlugin):
		super().__init__(mcdr_server)
		self.__plugin = plugin
		self.__logger_for_plugin = None  # type: Optional[MCDReforgedLogger]

	@property
	def logger(self) -> MCDReforgedLogger:
		if self.__logger_for_plugin is None:
			try:
				self.__logger_for_plugin = MCDReforgedLogger(self.__plugin.get_id())
			except:
				return super().logger
		return self.__logger_for_plugin

	def as_plugin_server_interface(self) -> Optional['PluginServerInterface']:
		return self

	# -----------------------
	#   Overwritten methods
	# -----------------------

	def get_plugin_command_source(self) -> PluginCommandSource:
		return PluginCommandSource(self, self.__plugin)

	# ------------------------
	#     Plugin Registry
	# ------------------------

	def register_event_listener(self, event: Union[PluginEvent, str], callback: Callable, priority: Optional[int] = None) -> None:
		"""
		Register an event listener for the current plugin
		:param event: The id of the event, or a PluginEvent instance. It indicates the target event for the plugin to listen
		:param callback: The callback listener method for the event
		:param priority: The priority of the listener. It will be set to the default value 1000 if it's not specified
		"""
		if priority is None:
			priority = DEFAULT_LISTENER_PRIORITY
		if isinstance(event, str):
			event = LiteralEvent(event_id=event)
		self.__plugin.register_event_listener(event, EventListener(self.__plugin, callback, priority))

	def register_command(self, root_node: Literal) -> None:
		"""
		Register an event listener for the current plugin
		:param root_node: the root node of your command tree. It should be a Literal node
		"""
		self.__plugin.register_command(root_node)

	def register_help_message(self, prefix: str, message: Union[MessageText, TranslationKeyDictRich], permission: int = PermissionLevel.MINIMUM_LEVEL) -> None:
		"""
		Register a help message for the current plugin, which is used in !!help command
		:param prefix: The help command of your plugin. When player click on the displayed message it will suggest this
		prefix parameter to the player. It's recommend to set it to the entry command of your plugin
		:param message: A neat command description. It can be a str or a RText. Also it can be a dict maps from language to description message
		:param permission: The minimum permission level for the user to see this help message. With default, anyone
		can see this message
		"""
		self.__plugin.register_help_message(HelpMessage(self.__plugin, prefix, message, permission))

	def register_translation(self, language: str, mapping: TranslationKeyDictNested) -> None:
		"""
		Register a translation mapping for a specific language for the current plugin
		:param language: The language of this translation
		:param mapping: A dict which maps translation keys into translated text. The translation key could be expressed as node name which under root node or the path of a nested multi-level nodes
		"""
		self.__plugin.register_translation(language, mapping)

	# ------------------------
	#      Plugin Utils
	# ------------------------

	def get_self_metadata(self) -> Metadata:
		"""
		Return the metadata of the plugin itself
		"""
		return self.__plugin.get_metadata()

	def get_data_folder(self) -> str:
		"""
		Return a unified data directory path for the current plugin
		The path of the folder will be "config/plugin_id/" where "plugin_id" is the id of the current plugin
		If the directory does not exist, create it
		:return: The path to the data directory
		"""
		plugin_data_folder = os.path.join(plugin_constant.PLUGIN_CONFIG_DIRECTORY, self.__plugin.get_id())
		if not os.path.isdir(plugin_data_folder):
			os.makedirs(plugin_data_folder)
		return plugin_data_folder

	def open_bundled_file(self, related_file_path: str) -> IO[bytes]:
		"""
		Open a file inside the plugin with readonly binary mode
		:param related_file_path: The related file path in your plugin to the file you want to open
		:return: A un-decoded bytes file-like object
		:raise: FileNotFoundError if the plugin is not a packed plugin (that is, a solo plugin)
		"""
		if not isinstance(self.__plugin, MultiFilePlugin):
			raise FileNotFoundError('Only packed plugin supported this API, found plugin type: {}'.format(self.__plugin.__class__))
		return self.__plugin.open_file(related_file_path)

	def load_config_simple(
			self, file_name: str = 'config.json', default_config: Optional = None, *,
			in_data_folder: bool = True, echo_in_console: bool = True, source_to_reply: Optional[CommandSource] = None,
			target_class: Optional[Type[SerializableType]] = None, encoding: str = 'utf8'
		) -> Union[dict, SerializableType]:
		"""
		A simple method to load a dict or Serializable type config from a json file
		Default config is supported. Missing key-values in the loaded config object will be filled using the default config
		:param file_name: The name of the config file. It can also be a path to the config file
		:param default_config: A dict contains the default config. It's required when the config file is missing,
		or exception will be risen. If target_class is given and default_config is missing, the default values in target_class
		will be used when the config file is missing
		:param in_data_folder: If True, the parent directory of file operating is the data folder of the plugin
		:param echo_in_console: If logging messages in console about config loading
		:param source_to_reply: The command source for replying logging messages
		:param target_class: A class derived from Serializable. When specified the loaded config data will be deserialized
		to a instance of target_class which will be returned as return value
		:param encoding: The encoding method to read the config file
		:return: A dict contains the loaded and processed config
		"""
		def log(msg):
			if isinstance(source_to_reply, CommandSource):
				source_to_reply.reply(msg)
			if echo_in_console:
				self.logger.info(msg)

		if target_class is not None:
			target_class: Serializable
			if default_config is None:
				default_config = target_class.get_default().serialize()
		config_file_path = os.path.join(self.get_data_folder(), file_name) if in_data_folder else file_name
		needs_save = False
		try:
			with open(config_file_path, encoding=encoding) as file_handler:
				read_data: dict = json.load(file_handler)
		except Exception as e:
			if default_config is not None:  # use default config
				result_config = default_config.copy()
			else:  # no default config and cannot read config file, raise the error
				raise
			needs_save = True
			log(self._mcdr_server.tr('server_interface.load_config_simple.failed', e))
		else:
			result_config = read_data
			if default_config is not None:
				# constructing the result config based on the given default config
				for key, value in default_config.items():
					if key in read_data:
						result_config[key] = read_data[key]
					else:
						result_config[key] = value
						log(self._mcdr_server.tr('server_interface.load_config_simple.key_missed', key, value))
						needs_save = True
			log(self._mcdr_server.tr('server_interface.load_config_simple.succeed'))
		if target_class is not None:
			try:
				result_config = target_class.deserialize(result_config)
			except Exception as e:
				result_config = target_class.get_default()
				needs_save = True
				log(self._mcdr_server.tr('server_interface.load_config_simple.failed', e))
		else:
			# remove unexpected keys
			if default_config is not None:
				for key in list(result_config.keys()):
					if key not in default_config:
						result_config.pop(key)
		if needs_save:
			self.save_config_simple(result_config, file_name=file_name, in_data_folder=in_data_folder)
		return result_config

	def save_config_simple(
			self, config: Union[dict, Serializable], file_name: str = 'config.json', *,
			in_data_folder: bool = True, encoding: str = 'utf8'
	) -> None:
		"""
		A simple method to save your dict or Serializable type config as a json file
		:param config: The config instance to be saved
		:param file_name: The name of the config file. It can also be a path to the config file
		:param in_data_folder: If True, the parent directory of file operating is the data folder of the plugin
		:param encoding: The encoding method to write the config file
		"""
		config_file_path = os.path.join(self.get_data_folder(), file_name) if in_data_folder else file_name
		if isinstance(config, Serializable):
			data = config.serialize()
		else:
			data = config
		target_folder = os.path.dirname(config_file_path)
		if len(target_folder) > 0 and not os.path.isdir(target_folder):
			os.makedirs(target_folder)
		with open(config_file_path, 'w', encoding=encoding) as file:
			# config file should be nicely readable, so here come the indent and non-ascii chars
			json.dump(data, file, indent=4, ensure_ascii=False)



import time
from contextlib import contextmanager
from typing import TYPE_CHECKING

from mcdreforged.executor.task_executor import TaskExecutor
from mcdreforged.executor.thread_executor import ThreadExecutor

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class WatchDog(ThreadExecutor):
	TASK_EXECUTOR_NO_RESPOND_THRESHOLD = 10  # seconds

	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		super().__init__(mcdr_server)
		self.__monitoring = False

	def is_monitoring(self):
		return self.__monitoring

	def pause(self):
		self.__monitoring = False

	def resume(self):
		self.__monitoring = True

	@contextmanager
	def pausing(self):
		self.pause()
		try:
			yield
		finally:
			self.resume()

	def check_task_executor_state(self):
		task_executor = self.mcdr_server.task_executor
		if task_executor.get_this_tick_time() > self.TASK_EXECUTOR_NO_RESPOND_THRESHOLD and self.__monitoring:
			plugin = self.mcdr_server.plugin_manager.get_current_running_plugin(thread=task_executor.get_thread())
			self.mcdr_server.logger.warning(self.mcdr_server.tr('watchdog.task_executor_no_response.line1', self.TASK_EXECUTOR_NO_RESPOND_THRESHOLD))
			self.mcdr_server.logger.warning(self.mcdr_server.tr('watchdog.task_executor_no_response.line2', plugin))
			self.mcdr_server.logger.warning(self.mcdr_server.tr('watchdog.task_executor_no_response.line3'))

			task_executor.soft_stop()  # Soft-stop it, in case it turns alive somehow
			task_executor.set_name(task_executor.get_name() + ' (no response)')

			new_executor = TaskExecutor(self.mcdr_server, previous_executor=task_executor)
			self.mcdr_server.set_task_executor(new_executor)
			new_executor.start()

	def start(self):
		super().start()
		self.__monitoring = True

	def tick(self):
		try:
			time.sleep(0.1)
			self.check_task_executor_state()
		except:
			self.mcdr_server.logger.exception('Error ticking watch dog')



"""
Permission control things
"""
from typing import Set

from mcdreforged.command.command_source import CommandSource
from mcdreforged.info_reactor.info import *
from mcdreforged.permission.permission_level import PermissionLevel, PermissionLevelItem, PermissionParam
from mcdreforged.utils import misc_util
from mcdreforged.utils.logger import DebugOption
from mcdreforged.utils.yaml_data_storage import YamlDataStorage

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


PERMISSION_FILE = 'permission.yml'
DEFAULT_PERMISSION_RESOURCE_PATH = 'resources/default_permission.yml'


class PermissionManager(YamlDataStorage):
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		super().__init__(mcdr_server.logger, PERMISSION_FILE, DEFAULT_PERMISSION_RESOURCE_PATH)
		self.mcdr_server = mcdr_server

	# --------------
	# File Operating
	# --------------

	def load_permission_file(self, *, allowed_missing_file: bool = True):
		"""
		Load the permission file from disk
		"""
		self._load_data(allowed_missing_file)

	def _pre_save(self, data: dict):
		# Deduplicate the permission data
		for key, value in data.items():
			if key in PermissionLevel.NAMES and isinstance(value, list):
				data[key] = misc_util.unique_list(data[key])
		# Change empty list to None for nicer look in the .yml file
		for key, value in data.items():
			if key in PermissionLevel.NAMES and value is not None and len(value) == 0:
				data[key] = None

	# ---------------------
	# Permission processing
	# ---------------------

	def get_default_permission_level(self) -> str:
		"""
		Return the default permission level
		"""
		return self._data['default_level']

	def set_default_permission_level(self, level: PermissionLevelItem):
		"""
		Set default permission level
		A message will be informed using server logger
		"""
		self._data['default_level'] = level.name
		self.save()
		self.mcdr_server.logger.info(self.mcdr_server.tr('permission_manager.set_default_permission_level.done', level.name))

	def get_permission_group_list(self, value: PermissionParam):
		"""
		Return the list of the player who has permission level <level>
		Example return value: ['Steve', 'Alex']

		:param value: a permission related object
		:rtype: list[str]
		"""
		level_name = PermissionLevel.from_value(value).name
		if self._data[level_name] is None:
			self._data[level_name] = []
		return self._data[level_name]

	def add_player(self, player: str, level_name: Optional[str] = None) -> int:
		"""
		Add a new player with permission level level_name
		If level_name is not set use default level
		Then save the permission data to file

		:param player: the name of the player
		:param level_name: the permission level name
		:return: the permission of the player after operation done
		"""
		if level_name is None:
			level_name = self.get_default_permission_level()
		PermissionLevel.from_value(level_name)  # validity check
		self.get_permission_group_list(level_name).append(player)
		self.mcdr_server.logger.debug('Added player {} with permission level {}'.format(player, level_name), option=DebugOption.PERMISSION)
		self.save()
		return PermissionLevel.from_value(level_name).level

	def remove_player(self, player: str):
		"""
		Remove a player from data, then save the permission data to file
		If the player has multiple permission level, remove them all
		And then save the permission data to file

		:param player: the name of the player
		"""
		while True:
			level = self.get_player_permission_level(player, auto_add=False)
			if level is None:
				break
			self.get_permission_group_list(level).remove(player)
		self.mcdr_server.logger.debug('Removed player {}'.format(player), option=DebugOption.PERMISSION)
		self.save()

	def set_permission_level(self, player: str, new_level: PermissionLevelItem):
		"""
		Set new permission level of the player
		Basically it will remove the player first, then add the player with given permission level
		Then save the permission data to file

		:param player: the name of the player
		:param new_level: the permission level name
		"""
		self.remove_player(player)
		self.add_player(player, new_level.name)
		self.mcdr_server.logger.info(self.mcdr_server.tr('permission_manager.set_permission_level.done', player, new_level.name))

	def touch_player(self, player: str):
		"""
		Add player if it's not in permission data

		:param player: the name of the player
		"""
		self.get_player_permission_level(player)

	def get_player_permission_level(self, player: str, *, auto_add: bool = True) -> Optional[int]:
		"""
		If the player is not in the permission data set its level to default_level,
		unless parameter auto_add is set to False, then it will return None
		If the player is in multiple permission level group it will return the highest one

		:param player: the name of the player
		:param auto_add: if it's True when player is invalid he will receive the default permission level
		:return the permission level from a player's name. If auto_add is False and player invalid return None
		"""
		for level_value in PermissionLevel.LEVELS[::-1]:  # high -> low
			if player in self.get_permission_group_list(level_value):
				return level_value
		else:
			if auto_add:
				return self.add_player(player)
			else:
				return None

	def get_permission(self, source: CommandSource) -> int:
		"""
		Gets called in CommandSource implementation
		"""
		if isinstance(source, ConsoleCommandSource):
			return PermissionLevel.CONSOLE_LEVEL
		elif isinstance(source, PlayerCommandSource):
			return self.get_player_permission_level(source.player)
		else:
			raise TypeError('Unknown type {} in get_permission'.format(type(source)))

	def get_players(self) -> Set[str]:
		players = set()
		for level_value in PermissionLevel.LEVELS:
			players.update(self.get_permission_group_list(level_value))
		return players
